import logging
import importlib
from pathlib import Path

from bbot import Preset
from ..test_step_2.module_tests.base import ModuleTestBase

log = logging.getLogger("bbot.test.modules")

module_tests_dir = Path(__file__).parent.parent / "test_step_2" / "module_tests"

_module_test_files = list(module_tests_dir.glob("test_module_*.py"))
_module_test_files.sort(key=lambda p: p.name)
module_test_files = [m.name.split("test_module_")[-1].split(".")[0] for m in _module_test_files]


def test__module__tests():
    preset = Preset()

    # make sure each module has a .py file
    for module_name in preset.module_loader.preloaded():
        module_name = module_name.lower()
        assert module_name in module_test_files, f'No test file found for module "{module_name}"'

    # make sure each test file has a test class
    for file in _module_test_files:
        module_name = file.stem
        import_path = f"bbot.test.test_step_2.module_tests.{module_name}"
        module_test_variables = importlib.import_module(import_path, "bbot")
        module_pass = False
        for var_name in dir(module_test_variables):
            if var_name.startswith("Test"):
                test_class = getattr(module_test_variables, var_name)
                if ModuleTestBase in getattr(test_class, "__mro__", ()):
                    module_pass = True
                    break
        assert module_pass, f"Couldn't find a test class for {module_name} in {file}"



import zipfile
import fnmatch

from bbot.modules.templates.github import github


class github_workflows(github):
    watched_events = ["CODE_REPOSITORY"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "safe", "code-enum"]
    meta = {
        "description": "Download a github repositories workflow logs and workflow artifacts",
        "created_date": "2024-04-29",
        "author": "@domwhewell-sage",
    }
    options = {"api_key": "", "num_logs": 1}
    options_desc = {
        "api_key": "Github token",
        "num_logs": "For each workflow fetch the last N successful runs logs (max 100)",
    }

    scope_distance_modifier = 2

    async def setup(self):
        self.num_logs = int(self.config.get("num_logs", 1))
        if self.num_logs > 100:
            self.log.error("num_logs option is capped at 100")
            return False
        self.output_dir = self.scan.home / "workflow_logs"
        self.helpers.mkdir(self.output_dir)
        return await super().setup()

    def _api_response_is_success(self, r):
        # we allow 404s because they're normal
        return r.is_success or getattr(r, "status_code", 0) == 404

    async def filter_event(self, event):
        if event.type == "CODE_REPOSITORY":
            if "git" not in event.tags and "github" not in event.data.get("url", ""):
                return False, "event is not a git repository"
        return True

    async def handle_event(self, event):
        repo_url = event.data.get("url")
        owner = repo_url.split("/")[-2]
        repo = repo_url.split("/")[-1]
        for workflow in await self.get_workflows(owner, repo):
            workflow_name = workflow.get("name")
            workflow_id = workflow.get("id")
            self.log.debug(f"Looking up runs for {workflow_name} in {owner}/{repo}")
            for run in await self.get_workflow_runs(owner, repo, workflow_id):
                run_id = run.get("id")
                workflow_url = f"https://github.com/{owner}/{repo}/actions/runs/{run_id}"
                self.log.debug(f"Downloading logs for {workflow_name}/{run_id} in {owner}/{repo}")
                for log in await self.download_run_logs(owner, repo, run_id):
                    logfile_event = self.make_event(
                        {
                            "path": str(log),
                            "description": f"Workflow run logs from {workflow_url}",
                        },
                        "FILESYSTEM",
                        tags=["textfile"],
                        parent=event,
                    )
                    await self.emit_event(
                        logfile_event,
                        context=f"{{module}} downloaded workflow run logs from {workflow_url} to {{event.type}}: {log}",
                    )
                artifacts = await self.get_run_artifacts(owner, repo, run_id)
                if artifacts:
                    for artifact in artifacts:
                        artifact_id = artifact.get("id")
                        artifact_name = artifact.get("name")
                        expired = artifact.get("expired")
                        if not expired:
                            filepath = await self.download_run_artifacts(owner, repo, artifact_id, artifact_name)
                            if filepath:
                                artifact_event = self.make_event(
                                    {
                                        "path": str(filepath),
                                        "description": f"Workflow run artifact from {workflow_url}",
                                    },
                                    "FILESYSTEM",
                                    tags=["zipfile"],
                                    parent=event,
                                )
                                await self.emit_event(
                                    artifact_event,
                                    context=f"{{module}} downloaded workflow run artifact from {workflow_url} to {{event.type}}: {filepath}",
                                )

    async def get_workflows(self, owner, repo):
        workflows = []
        url = f"{self.base_url}/repos/{owner}/{repo}/actions/workflows?per_page=100&page=" + "{page}"
        agen = self.api_page_iter(url, _json=False)
        try:
            async for r in agen:
                if r is None:
                    break
                status_code = getattr(r, "status_code", 0)
                if status_code == 403:
                    self.warning("Github is rate-limiting us (HTTP status: 403)")
                    break
                if status_code != 200:
                    break
                try:
                    j = r.json().get("workflows", [])
                except Exception as e:
                    self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                    break
                if not j:
                    break
                for item in j:
                    workflows.append(item)
        finally:
            await agen.aclose()
        return workflows

    async def get_workflow_runs(self, owner, repo, workflow_id):
        runs = []
        url = f"{self.base_url}/repos/{owner}/{repo}/actions/workflows/{workflow_id}/runs?status=success&per_page={self.num_logs}"
        r = await self.api_request(url)
        if r is None:
            return runs
        status_code = getattr(r, "status_code", 0)
        if status_code == 403:
            self.warning("Github is rate-limiting us (HTTP status: 403)")
            return runs
        if status_code != 200:
            return runs
        try:
            j = r.json().get("workflow_runs", [])
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return runs
        if not j:
            return runs
        for item in j:
            runs.append(item)
        return runs

    async def download_run_logs(self, owner, repo, run_id):
        folder = self.output_dir / owner / repo
        self.helpers.mkdir(folder)
        filename = f"run_{run_id}.zip"
        file_destination = folder / filename
        try:
            await self.helpers.download(
                f"{self.base_url}/repos/{owner}/{repo}/actions/runs/{run_id}/logs",
                filename=file_destination,
                headers=self.headers,
                raise_error=True,
                warn=False,
            )
            self.info(f"Downloaded logs for {owner}/{repo}/{run_id} to {file_destination}")
        except Exception as e:
            file_destination = None
            response = getattr(e, "response", None)
            status_code = getattr(response, "status_code", 0)
            if status_code == 403:
                self.warning(
                    f"The current access key does not have access to workflow {owner}/{repo}/{run_id} (status: {status_code})"
                )
            else:
                self.info(
                    f"The logs for {owner}/{repo}/{run_id} have expired and are no longer available (status: {status_code})"
                )
        # Secrets are duplicated in the individual workflow steps so just extract the main log files from the top folder
        if file_destination:
            main_logs = []
            with zipfile.ZipFile(file_destination, "r") as logzip:
                for name in logzip.namelist():
                    if fnmatch.fnmatch(name, "*.txt") and "/" not in name:
                        logzip.extract(name, folder)
                        main_logs.append(folder / name)
            return main_logs
        else:
            return []

    async def get_run_artifacts(self, owner, repo, run_id):
        artifacts = []
        url = f"{self.base_url}/repos/{owner}/{repo}/actions/runs/{run_id}/artifacts"
        r = await self.api_request(url)
        if r is None:
            return artifacts
        status_code = getattr(r, "status_code", 0)
        if status_code == 403:
            self.warning("Github is rate-limiting us (HTTP status: 403)")
            return artifacts
        if status_code != 200:
            return artifacts
        try:
            j = r.json().get("artifacts", [])
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return artifacts
        if not j:
            return artifacts
        for item in j:
            artifacts.append(item)
        return artifacts

    async def download_run_artifacts(self, owner, repo, artifact_id, artifact_name):
        folder = self.output_dir / owner / repo
        self.helpers.mkdir(folder)
        file_destination = folder / artifact_name
        try:
            await self.helpers.download(
                f"{self.base_url}/repos/{owner}/{repo}/actions/artifacts/{artifact_id}/zip",
                filename=file_destination,
                headers=self.headers,
                raise_error=True,
                warn=False,
            )
            self.info(
                f"Downloaded workflow artifact {owner}/{repo}/{artifact_id}/{artifact_name} to {file_destination}"
            )
        except Exception as e:
            file_destination = None
            response = getattr(e, "response", None)
            status_code = getattr(response, "status_code", 0)
            if status_code == 403:
                self.warning(
                    f"The current access key does not have access to workflow artifacts {owner}/{repo}/{artifact_id} (status: {status_code})"
                )
        return file_destination



from bbot.modules.templates.postman import postman


class postman(postman):
    watched_events = ["ORG_STUB", "SOCIAL"]
    produced_events = ["CODE_REPOSITORY"]
    flags = ["passive", "subdomain-enum", "safe", "code-enum"]
    meta = {
        "description": "Query Postman's API for related workspaces, collections, requests and download them",
        "created_date": "2024-09-07",
        "author": "@domwhewell-sage",
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "Postman API Key"}
    reject_wildcards = False

    async def handle_event(self, event):
        # Handle postman profile
        if event.type == "SOCIAL":
            owner = event.data.get("profile_name", "")
            in_scope_workspaces = await self.process_workspaces(user=owner)
        elif event.type == "ORG_STUB":
            owner = event.data
            in_scope_workspaces = await self.process_workspaces(org=owner)
        if in_scope_workspaces:
            for workspace in in_scope_workspaces:
                repo_url = workspace["url"]
                repo_name = workspace["repo_name"]
                if event.type == "SOCIAL":
                    context = f'{{module}} searched postman.com for workspaces belonging to "{owner}" and found "{repo_name}" at {{event.type}}: {repo_url}'
                elif event.type == "ORG_STUB":
                    context = f'{{module}} searched postman.com for "{owner}" and found matching workspace "{repo_name}" at {{event.type}}: {repo_url}'
                await self.emit_event(
                    {"url": repo_url},
                    "CODE_REPOSITORY",
                    tags="postman",
                    parent=event,
                    context=context,
                )

    async def process_workspaces(self, user=None, org=None):
        in_scope_workspaces = []
        owner = user or org
        if owner:
            self.verbose(f"Searching for postman workspaces, collections, requests for {owner}")
            for item in await self.query(owner):
                workspace = item["document"]
                slug = workspace["slug"]
                profile = workspace["publisherHandle"]
                repo_url = f"{self.html_url}/{profile}/{slug}"
                workspace_id = await self.get_workspace_id(repo_url)
                if (org and workspace_id) or (user and owner.lower() == profile.lower()):
                    self.verbose(f"Found workspace ID {workspace_id} for {repo_url}")
                    data = await self.request_workspace(workspace_id)
                    in_scope = await self.validate_workspace(
                        data["workspace"], data["environments"], data["collections"]
                    )
                    if in_scope:
                        in_scope_workspaces.append({"url": repo_url, "repo_name": slug})
                    else:
                        self.verbose(
                            f"Failed to validate {repo_url} is in our scope as it does not contain any in-scope dns_names / emails"
                        )
        return in_scope_workspaces

    async def query(self, query):
        def api_page_iter(url, page, page_size, offset, **kwargs):
            kwargs["json"]["body"]["from"] = offset
            return url, kwargs

        data = []
        url = f"{self.base_url}/ws/proxy"
        json = {
            "service": "search",
            "method": "POST",
            "path": "/search-all",
            "body": {
                "queryIndices": [
                    "collaboration.workspace",
                ],
                "queryText": self.helpers.quote(query),
                "size": 25,
                "from": 0,
                "clientTraceId": "",
                "requestOrigin": "srp",
                "mergeEntities": "true",
                "nonNestedRequests": "true",
                "domain": "public",
            },
        }

        agen = self.api_page_iter(
            url, page_size=25, method="POST", iter_key=api_page_iter, json=json, _json=False, headers=self.headers
        )
        async for r in agen:
            status_code = getattr(r, "status_code", 0)
            if status_code != 200:
                self.debug(f"Reached end of postman search results (url: {r.url}) with status code {status_code}")
                break
            try:
                data.extend(r.json().get("data", []))
            except Exception as e:
                self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                return None

        return data



from .base import ModuleTestBase


class TestBinaryEdge(ModuleTestBase):
    config_overrides = {"modules": {"binaryedge": {"api_key": "asdf"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.binaryedge.io/v2/query/domains/subdomain/blacklanternsecurity.com",
            match_headers={"X-Key": "asdf"},
            json={
                "query": "blacklanternsecurity.com",
                "page": 1,
                "pagesize": 100,
                "total": 1,
                "events": [
                    "asdf.blacklanternsecurity.com",
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.binaryedge.io/v2/user/subscription",
            match_headers={"X-Key": "asdf"},
            json={
                "subscription": {"name": "Free"},
                "end_date": "2023-06-17",
                "requests_left": 249,
                "requests_plan": 250,
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



import asyncio
import regex as re
from . import misc


class RegexHelper:
    """
    Class for misc CPU-intensive regex operations

    Offloads regex processing to other CPU cores via GIL release + thread pool

    For quick, one-off regexes, you don't need to use this helper.
    Only use this helper if you're searching large bodies of text
    or if your regex is CPU-intensive
    """

    def __init__(self, parent_helper):
        self.parent_helper = parent_helper

    def ensure_compiled_regex(self, r):
        """
        Make sure a regex has been compiled
        """
        if not isinstance(r, re.Pattern):
            raise ValueError("Regex must be compiled first!")

    def compile(self, *args, **kwargs):
        return re.compile(*args, **kwargs)

    async def search(self, compiled_regex, *args, **kwargs):
        self.ensure_compiled_regex(compiled_regex)
        return await self.parent_helper.run_in_executor(compiled_regex.search, *args, **kwargs)

    async def sub(self, compiled_regex, *args, **kwargs):
        self.ensure_compiled_regex(compiled_regex)
        return await self.parent_helper.run_in_executor(compiled_regex.sub, *args, **kwargs)

    async def findall(self, compiled_regex, *args, **kwargs):
        self.ensure_compiled_regex(compiled_regex)
        return await self.parent_helper.run_in_executor(compiled_regex.findall, *args, **kwargs)

    async def findall_multi(self, compiled_regexes, *args, threads=10, **kwargs):
        """
        Same as findall() but with multiple regexes
        """
        if not isinstance(compiled_regexes, dict):
            raise ValueError('compiled_regexes must be a dictionary like this: {"regex_name": <compiled_regex>}')
        for v in compiled_regexes.values():
            self.ensure_compiled_regex(v)

        tasks = {}

        def new_task(regex_name, r):
            task = self.parent_helper.run_in_executor(r.findall, *args, **kwargs)
            tasks[task] = regex_name

        compiled_regexes = dict(compiled_regexes)
        for _ in range(threads):  # Start initial batch of tasks
            if compiled_regexes:  # Ensure there are args to process
                new_task(*compiled_regexes.popitem())

        while tasks:  # While there are tasks pending
            # Wait for the first task to complete
            done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

            for task in done:
                result = task.result()
                regex_name = tasks.pop(task)
                yield (regex_name, result)

                if compiled_regexes:  # Start a new task for each one completed, if URLs remain
                    new_task(*compiled_regexes.popitem())

    async def finditer(self, compiled_regex, *args, **kwargs):
        self.ensure_compiled_regex(compiled_regex)
        return await self.parent_helper.run_in_executor(self._finditer, compiled_regex, *args, **kwargs)

    async def finditer_multi(self, compiled_regexes, *args, **kwargs):
        """
        Same as finditer() but with multiple regexes
        """
        for r in compiled_regexes:
            self.ensure_compiled_regex(r)
        return await self.parent_helper.run_in_executor(self._finditer_multi, compiled_regexes, *args, **kwargs)

    def _finditer_multi(self, compiled_regexes, *args, **kwargs):
        matches = []
        for r in compiled_regexes:
            for m in r.finditer(*args, **kwargs):
                matches.append(m)
        return matches

    def _finditer(self, compiled_regex, *args, **kwargs):
        return list(compiled_regex.finditer(*args, **kwargs))

    async def extract_params_html(self, *args, **kwargs):
        return await self.parent_helper.run_in_executor(misc.extract_params_html, *args, **kwargs)

    async def extract_emails(self, *args, **kwargs):
        return await self.parent_helper.run_in_executor(misc.extract_emails, *args, **kwargs)

    async def search_dict_values(self, *args, **kwargs):
        def _search_dict_values(*_args, **_kwargs):
            return list(misc.search_dict_values(*_args, **_kwargs))

        return await self.parent_helper.run_in_executor(_search_dict_values, *args, **kwargs)

    async def recursive_decode(self, *args, **kwargs):
        return await self.parent_helper.run_in_executor(misc.recursive_decode, *args, **kwargs)



from sys import executable
from urllib.parse import urlparse

from bbot.modules.base import BaseModule


class telerik(BaseModule):
    """
    Test for endpoints associated with Telerik.Web.UI.dll

    Telerik.Web.UI.WebResource.axd (CVE-2017-11317)
    Telerik.Web.UI.DialogHandler.aspx (CVE-2017-9248)
    Telerik.Web.UI.SpellCheckHandler.axd (associated with CVE-2017-9248)
    ChartImage.axd (CVE-2019-19790)

    For the Telerik Report Server vulnerability (CVE-2024-4358) Use the Nuclei Template: (https://github.com/projectdiscovery/nuclei-templates/blob/main/http/cves/2024/CVE-2024-4358.yaml)

    With exploit_RAU_crypto enabled, the module will attempt to exploit CVE-2017-11317. THIS WILL UPLOAD A (benign) FILE IF SUCCESSFUL.

    Will dedupe to host by default (running against first received URL). With include_subdirs enabled, will run against every directory.
    """

    watched_events = ["URL", "HTTP_RESPONSE"]
    produced_events = ["VULNERABILITY", "FINDING"]
    flags = ["active", "aggressive", "web-thorough"]
    meta = {
        "description": "Scan for critical Telerik vulnerabilities",
        "created_date": "2022-04-10",
        "author": "@liquidsec",
    }

    telerikVersions = [
        "2007.1423",
        "2007.1521",
        "2007.1626",
        "2007.2918",
        "2007.21010",
        "2007.21107",
        "2007.31218",
        "2007.31314",
        "2007.31425",
        "2008.1415",
        "2008.1515",
        "2008.1619",
        "2008.2723",
        "2008.2826",
        "2008.21001",
        "2008.31105",
        "2008.31125",
        "2008.31314",
        "2009.1311",
        "2009.1402",
        "2009.1527",
        "2009.2701",
        "2009.2826",
        "2009.31103",
        "2009.31208",
        "2009.31314",
        "2010.1309",
        "2010.1415",
        "2010.1519",
        "2010.2713",
        "2010.2826",
        "2010.2929",
        "2010.31109",
        "2010.31215",
        "2010.31317",
        "2011.1315",
        "2011.1413",
        "2011.1519",
        "2011.2712",
        "2011.2915",
        "2011.31115",
        "2011.3.1305",
        "2012.1.215",
        "2012.1.411",
        "2012.2.607",
        "2012.2.724",
        "2012.2.912",
        "2012.3.1016",
        "2012.3.1205",
        "2012.3.1308",
        "2013.1.220",
        "2013.1.403",
        "2013.1.417",
        "2013.2.611",
        "2013.2.717",
        "2013.3.1015",
        "2013.3.1114",
        "2013.3.1324",
        "2014.1.225",
        "2014.1.403",
        "2014.2.618",
        "2014.2.724",
        "2014.3.1024",
        "2015.1.204",
        "2015.1.225",
        "2015.2.604",
        "2015.2.623",
        "2015.2.729",
        "2015.2.826",
        "2015.3.930",
        "2015.3.1111",
        "2016.1.113",
        "2016.1.225",
        "2016.2.504",
        "2016.2.607",
        "2016.3.914",
        "2016.3.1018",
        "2016.3.1027",
        "2016.1.1213",
        "2017.1.118",
        "2017.1.228",
        "2017.2.503",
        "2017.2.621",
        "2017.2.711",
        "2017.3.913",
    ]

    DialogHandlerUrls = [
        "Telerik.Web.UI.DialogHandler.aspx",
        "Telerik.Web.UI.DialogHandler.axd",
        "Admin/ServerSide/Telerik.Web.UI.DialogHandler.aspx",
        "App_Master/Telerik.Web.UI.DialogHandler.aspx",
        "AsiCommon/Controls/ContentManagement/ContentDesigner/Telerik.Web.UI.DialogHandler.aspx",
        "cms/portlets/telerik.web.ui.dialoghandler.aspx",
        "common/admin/Calendar/Telerik.Web.UI.DialogHandler.aspx",
        "common/admin/Jobs2/Telerik.Web.UI.DialogHandler.aspx",
        "common/admin/PhotoGallery2/Telerik.Web.UI.DialogHandler.aspx",
        "dashboard/UserControl/CMS/Page/Telerik.Web.UI.DialogHandler.aspx",
        "DesktopModule/UIQuestionControls/UIAskQuestion/Telerik.Web.UI.DialogHandler.aspx",
        "Desktopmodules/Admin/dnnWerk.Users/DialogHandler.aspx",
        "DesktopModules/Admin/RadEditorProvider/DialogHandler.aspx",
        "desktopmodules/base/editcontrols/telerik.web.ui.dialoghandler.aspx",
        "desktopmodules/dnnwerk.radeditorprovider/dialoghandler.aspx",
        "DesktopModules/RadEditorProvider/telerik.web.ui.dialoghandler.aspx",
        "desktopmodules/tcmodules/tccategory/telerik.web.ui.dialoghandler.aspx",
        "desktopmodules/telerikwebui/radeditorprovider/telerik.web.ui.dialoghandler.aspx",
        "DesktopModules/TNComments/Telerik.Web.UI.DialogHandler.aspx",
        "dotnetnuke/DesktopModules/Admin/RadEditorProvider/DialogHandler.aspx",
        "Modules/CMS/Telerik.Web.UI.DialogHandler.aspx",
        "modules/shop/manage/telerik.web.ui.dialoghandler.aspx",
        "portal/channels/fa/Cms_HtmlText_Manage/Telerik.Web.UI.DialogHandler.aspx",
        "providers/htmleditorproviders/telerik/telerik.web.ui.dialoghandler.aspx",
        "Resources/Telerik.Web.UI.DialogHandler.aspx",
        "sitecore/shell/applications/contentmanager/telerik.web.ui.dialoghandler.aspx",
        "sitecore/shell/Controls/RichTextEditor/Telerik.Web.UI.DialogHandler.aspx",
        "Sitefinity/ControlTemplates/Blogs/Telerik.Web.UI.DialogHandler.aspx",
        "SiteTemplates/Telerik.Web.UI.DialogHandler.aspx",
        "static/usercontrols/Telerik.Web.UI.DialogHandler.aspx",
        "system/providers/htmleditor/Telerik.Web.UI.DialogHandler.aspx",
        "WebUIDialogs/Telerik.Web.UI.DialogHandler.aspx",
    ]

    RAUConfirmed = []

    options = {"exploit_RAU_crypto": False, "include_subdirs": False}
    options_desc = {
        "exploit_RAU_crypto": "Attempt to confirm any RAU AXD detections are vulnerable",
        "include_subdirs": "Include subdirectories in the scan (off by default)",  # will create many finding events if used in conjunction with web spider or ffuf
    }

    in_scope_only = True

    deps_pip = ["pycryptodome~=3.17"]

    deps_ansible = [
        {"name": "Create telerik dir", "file": {"state": "directory", "path": "#{BBOT_TOOLS}/telerik/"}},
        {"file": {"state": "touch", "path": "#{BBOT_TOOLS}/telerik/testfile.txt"}},
        {
            "name": "Download RAU_crypto",
            "unarchive": {
                "src": "https://github.com/bao7uo/RAU_crypto/archive/refs/heads/master.zip",
                "include": "RAU_crypto-master/RAU_crypto.py",
                "dest": "#{BBOT_TOOLS}/telerik/",
                "remote_src": True,
            },
        },
    ]

    _module_threads = 5

    @staticmethod
    def normalize_url(url):
        return str(url.rstrip("/") + "/").lower()

    def _incoming_dedup_hash(self, event):
        if event.type == "URL":
            if self.config.get("include_subdirs") is True:
                return hash(f"{event.type}{self.normalize_url(event.data)}")
            else:
                return hash(f"{event.type}{event.netloc}")
        else:  # HTTP_RESPONSE
            return hash(f"{event.type}{event.data['url']}")

    async def handle_event(self, event):
        if event.type == "URL":
            if self.config.get("include_subdirs"):
                base_url = self.normalize_url(event.data)  # Use the entire URL including subdirectories

            else:
                base_url = f"{event.parsed_url.scheme}://{event.parsed_url.netloc}/"  # path will be omitted

            # Check for RAU AXD Handler
            webresource = "Telerik.Web.UI.WebResource.axd?type=rau"
            result, _ = await self.test_detector(base_url, webresource)
            if result:
                if "RadAsyncUpload handler is registered successfully" in result.text:
                    self.verbose("Detected Telerik instance (Telerik.Web.UI.WebResource.axd?type=rau)")

                    probe_data = {
                        "rauPostData": (
                            None,
                            "mQheol55IDiQWWSxl+Atkc68JXWUJ6QSirwLhEwleMiw3vN4cwABE74V2fWsLGg8CFXHOP6np90M+sLrLDqFACGNvonxmgT8aBsTZPWbXErewMGNWBP34aX0DmMvXVyTEpQ6FkFhZi19cTtdYfRLI8Uc04uNSsdWnltDMQ2CX/sSLOXUFNnZdAwAXgUuprYhU28Zwh/GdgYh447ksXfAC2fuPqEJqKDDwBlltxsS/zSq8ipIg326ymB2dmOpH/P3hcAmTKOyzB0dW6a6pmJvqNVU+50DlrUC00RbBbTJwlV6Xm4s4XTvgXLvMQ6czz2OAYY18HI+HYX5uvajctj/25UR8edwu68ZCgedsD7EZHRSSthjxohxfAyrfshjcu1LnhCEd0ClowKxBS4eiaLxVxhJAdB7XcbbXxIS9WWKa7gtRMNc/jUAOlIpvOZ3N+bOQ6rsNMHv7TZk1g0bxPl99yBn9qvtAwDMNPDoADxoBSisAkIIl9mImKv7y7nAiKoj7ukApdu5XQuVo10SxwkLkqHcvEEgjxTrOlCbEbxK2/du9TgXxD9iqKyaPLHPzNZsnzCsG6qNXv0fNkeASP9tZAyvi/y1eLrpScE+J7blfT+kBkGPTTFc6Z4z6lN7GqSHofq/CDHC2S2+qdoRdC3C25V74j+Ae6MkpSfqYx4KZYNtxBAxjf9Uf3JVSiZh3X2W/7aFeimFft0h/liybSjJTzO+AwNJluI4kXqemFoHnjVFfUQViaIuk4UP0D861kCU6KIGLZLpOaa0g0KM8hmu3OjwVOy8QVXYtbx5lOmSX9h3imRzMDFRTXK25YpUJgD0/LFMgCeZLA8SCYzkThyN2d8f8n5l8iOScR47o8i8sqCp/fd3JTogSbwD7LxnHudpiw2W/OfpMGipgc6loQFoX4klQaYwKkA4w+GUzahfAJmIiukZuTLOPCPQvX4wKtLqw1YiHtuaLHvLYq2/F66QQXNrZ4SucUNED0p5TUVTvHGUbuA0zxAyYSfYVgTNZjXGguQBY7DsN1SkpCa/ltvIiGtCbHQR86OrvjJMACe0wdpMCqEg7JiGym3RrLqvmjpS&sbZRwxJ96gmXFBSbSvT0ve7jpvDoieqd6RbG+GIP0H7sO5/0ZnvheosB9jQAifuMabY7lW4UzZgr5o2iqE0tBl4SGhfWyYW7iCFXnd3aIuCnUvhT58Rp8g7kGkA/eU/s68E66KOBXNuBnokZR9cIsjE0Tt3Jfxrk018+CmVcXpjXp/RmhRwCJTgEAXQuNplb/KdkLxqDn519iRtbiU6aLZX8YctdFQBqyKVgkk8WYXxcXQ8wYnxtpEtGuBcsndUi1iPp4Od8rYY1HPWg+FIquW17YPHjfP4gO4dhZe4sd7gH0ARyGDjiYVj7ODDE0wGmwmFVdQTrDX5AaxKuJy0NbQ==",
                        ),
                        "file": ("blob", b"e1daf48a", "application/octet-stream"),
                        "fileName": (None, "df8dbc7a"),
                        "contentType": (None, "text/html"),
                        "lastModifiedDate": (None, "2020-01-02T08:02:01.067Z"),
                        "metadata": (
                            None,
                            '{"TotalChunks":1,"ChunkIndex":0,"TotalFileSize":1,"UploadID":"3ea7b19db6c5.txt"}',
                        ),
                    }

                    version = "unknown"
                    verbose_errors = False
                    # send probe
                    probe_response = await self.helpers.request(
                        f"{event.data}{webresource}", method="POST", files=probe_data
                    )

                    if probe_response:
                        if "Exception Details: " in probe_response.text:
                            verbose_errors = True
                            if (
                                "Telerik.Web.UI.CryptoExceptionThrower.ThrowGenericCryptoException"
                                in probe_response.text
                            ):
                                version = "Post-2020 (Encrypt-Then-Mac Enabled, with Generic Crypto Failure Message)"
                            elif "Padding is invalid and cannot be removed" in probe_response.text:
                                version = "<= 2019 (Either Pre-2017 (vulnerable), or 2017-2019 w/ Encrypt-Then-Mac)"

                    description = f"Telerik RAU AXD Handler detected. Verbose Errors Enabled: [{str(verbose_errors)}] Version Guess: [{version}]"
                    await self.emit_event(
                        {"host": str(event.host), "url": f"{base_url}{webresource}", "description": description},
                        "FINDING",
                        event,
                        context=f"{{module}} scanned {base_url} and identified {{event.type}}: Telerik RAU AXD Handler",
                    )
                    if self.config.get("exploit_RAU_crypto") is True:
                        if base_url not in self.RAUConfirmed:
                            self.RAUConfirmed.append(base_url)
                            root_tool_path = self.scan.helpers.tools_dir / "telerik"
                            self.debug(root_tool_path)

                            for version in self.telerikVersions:
                                command = [
                                    executable,
                                    str(root_tool_path / "RAU_crypto-master/RAU_crypto.py"),
                                    "-P",
                                    "C:\\\\Windows\\\\Temp",
                                    version,
                                    str(root_tool_path / "testfile.txt"),
                                    result.url,
                                ]
                                output = await self.run_process(command)
                                description = f"[CVE-2017-11317] [{str(version)}] {webresource}"
                                if "fileInfo" in output.stdout:
                                    self.debug(f"Confirmed Vulnerable Telerik (version: {str(version)}")
                                    await self.emit_event(
                                        {
                                            "severity": "CRITICAL",
                                            "description": description,
                                            "host": str(event.host),
                                            "url": f"{base_url}{webresource}",
                                        },
                                        "VULNERABILITY",
                                        event,
                                        context=f"{{module}} scanned {base_url} and identified critical {{event.type}}: {description}",
                                    )
                                    break

            urls = {}
            for dh in self.DialogHandlerUrls:
                url = self.create_url(base_url, f"{dh}?dp=1")
                urls[url] = dh

            gen = self.helpers.request_batch(list(urls))
            fail_count = 0
            async for url, response in gen:
                # cancel if we run into timeouts etc.
                if response is None:
                    fail_count += 1

                    # tolerate some random errors
                    if fail_count < 2:
                        continue
                    self.debug(f"Cancelling run against {base_url} due to failed request")
                    await gen.aclose()
                else:
                    if "Cannot deserialize dialog parameters" in response.text:
                        self.debug(f"Detected Telerik UI instance ({dh})")
                        description = "Telerik DialogHandler detected"
                        await self.emit_event(
                            {"host": str(event.host), "url": f"{base_url}{dh}", "description": description},
                            "FINDING",
                            event,
                        )
                        # Once we have a match we need to stop, because the basic handler (Telerik.Web.UI.DialogHandler.aspx) usually works with a path wildcard
                        await gen.aclose()

            spellcheckhandler = "Telerik.Web.UI.SpellCheckHandler.axd"
            result, _ = await self.test_detector(base_url, spellcheckhandler)
            status_code = getattr(result, "status_code", 0)
            # The standard behavior for the spellcheck handler without parameters is a 500
            if status_code == 500:
                # Sometimes webapps will just return 500 for everything, so rule out the false positive
                validate_result, _ = await self.test_detector(base_url, f"{self.helpers.rand_string()}.axd")
                self.debug(validate_result)
                validate_status_code = getattr(validate_result, "status_code", 0)
                if validate_status_code not in (0, 500):
                    self.debug("Detected Telerik UI instance (Telerik.Web.UI.SpellCheckHandler.axd)")
                    description = "Telerik SpellCheckHandler detected"
                    await self.emit_event(
                        {
                            "host": str(event.host),
                            "url": f"{base_url}{spellcheckhandler}",
                            "description": description,
                        },
                        "FINDING",
                        event,
                        context=f"{{module}} scanned {base_url} and identified {{event.type}}: Telerik SpellCheckHandler",
                    )

            chartimagehandler = "ChartImage.axd?ImageName=bqYXJAqm315eEd6b%2bY4%2bGqZpe7a1kY0e89gfXli%2bjFw%3d"
            result, _ = await self.test_detector(base_url, chartimagehandler)
            status_code = getattr(result, "status_code", 0)
            if status_code == 200:
                chartimagehandler_error = "ChartImage.axd?ImageName="
                result_error, _ = await self.test_detector(base_url, chartimagehandler_error)
                error_status_code = getattr(result_error, "status_code", 0)
                if error_status_code not in (0, 200):
                    await self.emit_event(
                        {
                            "host": str(event.host),
                            "url": f"{base_url}{chartimagehandler}",
                            "description": "Telerik ChartImage AXD Handler Detected",
                        },
                        "FINDING",
                        event,
                        context=f"{{module}} scanned {base_url} and identified {{event.type}}: Telerik ChartImage AXD Handler",
                    )

        elif event.type == "HTTP_RESPONSE":
            resp_body = event.data.get("body", None)
            url = event.data["url"]
            if resp_body:
                if '":{"SerializedParameters":"' in resp_body:
                    await self.emit_event(
                        {
                            "host": str(event.host),
                            "url": url,
                            "description": "Telerik DialogHandler [SerializedParameters] Detected in HTTP Response",
                        },
                        "FINDING",
                        event,
                        context="{module} searched HTTP_RESPONSE and identified {event.type}: Telerik ChartImage AXD Handler",
                    )
                elif '"_serializedConfiguration":"' in resp_body:
                    await self.emit_event(
                        {
                            "host": str(event.host),
                            "url": url,
                            "description": "Telerik AsyncUpload [serializedConfiguration] Detected in HTTP Response",
                        },
                        "FINDING",
                        event,
                        context="{module} searched HTTP_RESPONSE and identified {event.type}: Telerik AsyncUpload",
                    )

    def create_url(self, baseurl, detector):
        return f"{baseurl}{detector}"

    async def test_detector(self, baseurl, detector):
        result = None
        url = self.create_url(baseurl, detector)
        result = await self.helpers.request(url, timeout=self.scan.httpx_timeout)
        return result, detector

    async def filter_event(self, event):
        if event.type == "URL" and "endpoint" in event.tags:
            return False
        else:
            return True



import asyncio
import re
from werkzeug.wrappers import Response

from .base import ModuleTestBase


def extract_subdomain_tag(data):
    pattern = r"([a-z0-9]{4})\.fakedomain\.fakeinteractsh\.com"
    match = re.search(pattern, data)
    if match:
        return match.group(1)


class TestHost_Header(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "host_header"]

    fake_host = "fakedomain.fakeinteractsh.com"

    def request_handler(self, request):
        subdomain_tag = None
        subdomain_tag = extract_subdomain_tag(request.headers["Host"])

        # Standard (with reflection)
        if subdomain_tag:
            self.interactsh_mock_instance.mock_interaction(subdomain_tag)
            return Response(f"Alive, host is: {subdomain_tag}.{self.fake_host}", status=200)

        # Host Header Overrides
        subdomain_tag_overrides = extract_subdomain_tag(request.headers["X-Forwarded-For"])
        if subdomain_tag_overrides:
            return Response(f"Alive, host is: {subdomain_tag}.{self.fake_host}", status=200)

        return Response("Alive, host is: defaulthost.com", status=200)

    async def setup_before_prep(self, module_test):
        self.interactsh_mock_instance = module_test.mock_interactsh("host_header")
        module_test.monkeypatch.setattr(
            module_test.scan.helpers, "interactsh", lambda *args, **kwargs: self.interactsh_mock_instance
        )

    async def setup_after_prep(self, module_test):
        expect_args = re.compile("/")
        module_test.set_expect_requests_handler(expect_args=expect_args, request_handler=self.request_handler)

    def check(self, module_test, events):
        # We can't fully test all the use-cases because werkzeug abstracts away some of our RFC-violating tricks :/

        for e in events:
            assert any(
                e.type == "FINDING"
                and "Possible Host header injection. Injection technique: standard" in e.data["description"]
                for e in events
            ), "Failed to detect Possible Host Header Injection (standard)"
            assert any(
                e.type == "FINDING"
                and "Possible Host header injection. Injection technique: host override headers"
                in e.data["description"]
                for e in events
            ), "Failed to detect Possible Host Header Injection (host override headers)"
            assert any(
                e.type == "FINDING" and "Spoofed Host header (standard) [HTTP] interaction" in e.data["description"]
                for e in events
            ), "Failed to detect Spoofed Host header (standard) [HTTP] interaction"



# adapted from https://github.com/bugcrowd/HUNT

from bbot.modules.base import BaseModule

hunt_param_dict = {
    "Command Injection": [
        "daemon",
        "host",
        "upload",
        "dir",
        "execute",
        "download",
        "log",
        "ip",
        "cli",
        "cmd",
        "exec",
        "command",
        "func",
        "code",
        "update",
        "shell",
        "eval",
    ],
    "Debug": [
        "access",
        "admin",
        "dbg",
        "debug",
        "edit",
        "grant",
        "test",
        "alter",
        "clone",
        "create",
        "delete",
        "disable",
        "enable",
        "exec",
        "execute",
        "load",
        "make",
        "modify",
        "rename",
        "reset",
        "shell",
        "toggle",
        "adm",
        "root",
        "cfg",
        "config",
    ],
    "Directory Traversal": ["entry", "download", "attachment", "basepath", "path", "file", "source", "dest"],
    "Local File Include": [
        "file",
        "document",
        "folder",
        "root",
        "path",
        "pg",
        "style",
        "pdf",
        "template",
        "php_path",
        "doc",
        "lang",
        "include",
        "img",
        "view",
        "layout",
        "export",
        "log",
        "configFile",
        "stylesheet",
        "configFileUrl",
    ],
    "Insecure Direct Object Reference": [
        "id",
        "user",
        "account",
        "number",
        "order",
        "no",
        "doc",
        "key",
        "email",
        "group",
        "profile",
        "edit",
        "report",
        "docId",
        "accountId",
        "customerId",
        "reportId",
        "jobId",
        "sessionId",
        "api_key",
        "instance",
        "identifier",
        "access",
    ],
    "SQL Injection": [
        "id",
        "select",
        "report",
        "role",
        "update",
        "query",
        "user",
        "name",
        "sort",
        "where",
        "search",
        "params",
        "category",
        "process",
        "row",
        "view",
        "table",
        "from",
        "sel",
        "results",
        "sleep",
        "fetch",
        "order",
        "keyword",
        "column",
        "field",
        "delete",
        "string",
        "number",
        "filter",
        "limit",
        "offset",
        "item",
        "input",
        "date",
        "value",
        "orderBy",
        "groupBy",
        "pageNum",
        "pageSize",
        "tag",
        "author",
        "postId",
        "parentId",
        "d",
    ],
    "Server-side Request Forgery": [
        "dest",
        "redirect",
        "uri",
        "path",
        "continue",
        "url",
        "window",
        "next",
        "data",
        "reference",
        "site",
        "html",
        "val",
        "validate",
        "domain",
        "callback",
        "return",
        "page",
        "feed",
        "host",
        "port",
        "to",
        "out",
        "view",
        "dir",
        "show",
        "navigation",
        "open",
        "proxy",
        "target",
        "server",
        "domain",
        "connect",
        "fetch",
        "apiEndpoint",
    ],
    "Server-Side Template Injection": [
        "template",
        "preview",
        "id",
        "view",
        "activity",
        "name",
        "content",
        "redirect",
        "expression",
        "statement",
        "tpl",
        "render",
        "format",
        "engine",
    ],
    "XML external entity injection": [
        "xml",
        "dtd",
        "xsd",
        "xmlDoc",
        "xmlData",
        "entityType",
        "entity",
        "xmlUrl",
        "schema",
        "xmlFile",
        "xmlPath",
        "xmlSource",
        "xmlEndpoint",
        "xslt",
        "xmlConfig",
        "xmlCallback",
        "attributeName",
        "wsdl",
        "xmlDocUrl",
    ],
    "Insecure Cryptography": [
        "encrypted",
        "cipher",
        "iv",
        "checksum",
        "hash",
        "salt",
        "hmac",
        "secret",
        "key",
        "signatureAlgorithm",
        "keyId",
        "sharedSecret",
        "privateKeyId",
        "privateKey",
        "publicKey",
        "publicKeyId",
        "encryptedData",
        "encryptedMessage",
        "encryptedPayload",
        "encryptedFile",
        "cipherText",
        "cipherAlgorithm",
        "keySize",
        "keyPair",
        "keyDerivation",
        "encryptionMethod",
        "decryptionKey",
    ],
    "Unsafe Deserialization": [
        "serialized",
        "object",
        "dataObject",
        "serialization",
        "payload",
        "encoded",
        "marshalled",
        "pickled",
        "jsonData",
        "state",
        "sessionData",
        "cache",
        "tokenData",
        "serializedSession",
        "objectState",
        "jsonDataPayload",
    ],
}


class hunt(BaseModule):
    watched_events = ["WEB_PARAMETER"]
    produced_events = ["FINDING"]
    flags = ["active", "safe", "web-thorough"]
    meta = {
        "description": "Watch for commonly-exploitable HTTP parameters",
        "author": "@liquidsec",
        "created_date": "2022-07-20",
    }
    # accept all events regardless of scope distance
    scope_distance_modifier = None

    async def handle_event(self, event):
        p = event.data["name"]
        for k in hunt_param_dict.keys():
            if p.lower() in hunt_param_dict[k]:
                description = f"Found potential {k.upper()} parameter [{p}]"
                data = {"host": str(event.host), "description": description}
                url = event.data.get("url", "")
                if url:
                    data["url"] = url
                await self.emit_event(data, "FINDING", event)



import os
import asyncio
import aiosqlite
import multiprocessing
from pathlib import Path
from contextlib import suppress
from shutil import copyfile, copymode

from bbot.modules.base import BaseModule


class gowitness(BaseModule):
    watched_events = ["URL", "SOCIAL"]
    produced_events = ["WEBSCREENSHOT", "URL", "URL_UNVERIFIED", "TECHNOLOGY"]
    flags = ["active", "safe", "web-screenshots"]
    meta = {"description": "Take screenshots of webpages", "created_date": "2022-07-08", "author": "@TheTechromancer"}
    options = {
        "version": "2.4.2",
        "threads": 0,
        "timeout": 10,
        "resolution_x": 1440,
        "resolution_y": 900,
        "output_path": "",
        "social": False,
        "idle_timeout": 1800,
    }
    options_desc = {
        "version": "Gowitness version",
        "threads": "How many gowitness threads to spawn (default is number of CPUs x 2)",
        "timeout": "Preflight check timeout",
        "resolution_x": "Screenshot resolution x",
        "resolution_y": "Screenshot resolution y",
        "output_path": "Where to save screenshots",
        "social": "Whether to screenshot social media webpages",
        "idle_timeout": "Skip the current gowitness batch if it stalls for longer than this many seconds",
    }
    deps_common = ["chromium"]
    deps_pip = ["aiosqlite"]
    deps_ansible = [
        {
            "name": "Download gowitness",
            "get_url": {
                "url": "https://github.com/sensepost/gowitness/releases/download/#{BBOT_MODULES_GOWITNESS_VERSION}/gowitness-#{BBOT_MODULES_GOWITNESS_VERSION}-#{BBOT_OS_PLATFORM}-#{BBOT_CPU_ARCH}",
                "dest": "#{BBOT_TOOLS}/gowitness",
                "mode": "755",
            },
        },
    ]
    _batch_size = 100
    # gowitness accepts SOCIAL events up to distance 2, otherwise it is in-scope-only
    scope_distance_modifier = 2

    async def setup(self):
        num_cpus = multiprocessing.cpu_count()
        default_thread_count = min(20, num_cpus * 2)
        self.timeout = self.config.get("timeout", 10)
        self.idle_timeout = self.config.get("idle_timeout", 1800)
        self.threads = self.config.get("threads", 0)
        if not self.threads:
            self.threads = default_thread_count
        self.proxy = self.scan.web_config.get("http_proxy", "")
        self.resolution_x = self.config.get("resolution_x")
        self.resolution_y = self.config.get("resolution_y")
        self.visit_social = self.config.get("social", True)
        output_path = self.config.get("output_path")
        if output_path:
            self.base_path = Path(output_path) / "gowitness"
        else:
            self.base_path = self.scan.home / "gowitness"
        self.chrome_path = None
        custom_chrome_path = self.helpers.tools_dir / "chrome-linux" / "chrome"
        if custom_chrome_path.is_file():
            self.chrome_path = custom_chrome_path

        # fix ubuntu-specific sandbox bug
        chrome_devel_sandbox = self.helpers.tools_dir / "chrome-linux" / "chrome_sandbox"
        if chrome_devel_sandbox.is_file():
            os.environ["CHROME_DEVEL_SANDBOX"] = str(chrome_devel_sandbox)

        # make sure we have a working chrome install
        chrome_test_pass = False
        for binary in ("chrome", "chromium", "chromium-browser", custom_chrome_path):
            binary_path = self.helpers.which(binary)
            if binary_path and Path(binary_path).is_file():
                chrome_test_proc = await self.run_process([binary_path, "--version"])
                if getattr(chrome_test_proc, "returncode", 1) == 0:
                    self.verbose(f"Found chrome executable at {binary_path}")
                    chrome_test_pass = True
                    break
        if not chrome_test_pass:
            return False, "Failed to set up Google chrome. Please install manually or try again with --force-deps."

        self.db_path = self.base_path / "gowitness.sqlite3"
        self.screenshot_path = self.base_path / "screenshots"
        self.command = self.construct_command()
        self.prepped = False
        self.screenshots_taken = {}
        self.connections_logged = set()
        self.technologies_found = set()
        return True

    def prep(self):
        if not self.prepped:
            self.helpers.mkdir(self.screenshot_path)
            self.db_path.touch()
            with suppress(Exception):
                copyfile(self.helpers.tools_dir / "gowitness", self.base_path / "gowitness")
                copymode(self.helpers.tools_dir / "gowitness", self.base_path / "gowitness")
            self.prepped = True

    async def filter_event(self, event):
        # Ignore URLs that are redirects
        if any(t.startswith("status-30") for t in event.tags):
            return False, "URL is a redirect"
        # ignore events from self
        if event.type == "URL" and event.module == self:
            return False, "event is from self"
        if event.type == "SOCIAL":
            if not self.visit_social:
                return False, "visit_social=False"
        else:
            # Accept out-of-scope SOCIAL pages, but not URLs
            if event.scope_distance > 0:
                return False, "event is not in-scope"
        return True

    async def handle_batch(self, *events):
        self.prep()
        event_dict = {}
        for e in events:
            key = e.data
            if e.type == "SOCIAL":
                key = e.data["url"]
            event_dict[key] = e
        stdin = "\n".join(list(event_dict))

        try:
            async for line in self.run_process_live(self.command, input=stdin, idle_timeout=self.idle_timeout):
                self.debug(line)
        except asyncio.exceptions.TimeoutError:
            urls_str = ",".join(event_dict)
            self.warning(f"Gowitness timed out while visiting the following URLs: {urls_str}", trace=False)
            return

        # emit web screenshots
        new_screenshots = await self.get_new_screenshots()
        for filename, screenshot in new_screenshots.items():
            url = screenshot["url"]
            final_url = screenshot["final_url"]
            filename = self.screenshot_path / screenshot["filename"]
            filename = filename.relative_to(self.scan.home)
            # NOTE: this prevents long filenames from causing problems in BBOT, but gowitness will still fail to save it.
            filename = self.helpers.truncate_filename(filename)
            webscreenshot_data = {"path": str(filename), "url": final_url}
            parent_event = event_dict[url]
            await self.emit_event(
                webscreenshot_data,
                "WEBSCREENSHOT",
                parent=parent_event,
                context=f"{{module}} visited {final_url} and saved {{event.type}} to {filename}",
            )

        # emit URLs
        new_network_logs = await self.get_new_network_logs()
        for url, row in new_network_logs.items():
            ip = row["ip"]
            status_code = row["status_code"]
            tags = [f"status-{status_code}", f"ip-{ip}", "spider-danger"]

            _id = row["url_id"]
            parent_url = self.screenshots_taken[_id]
            parent_event = event_dict[parent_url]
            if url and url.startswith("http"):
                await self.emit_event(
                    url,
                    "URL_UNVERIFIED",
                    parent=parent_event,
                    tags=tags,
                    context=f"{{module}} visited {{event.type}}: {url}",
                )

        # emit technologies
        new_technologies = await self.get_new_technologies()
        for row in new_technologies.values():
            parent_id = row["url_id"]
            parent_url = self.screenshots_taken[parent_id]
            parent_event = event_dict[parent_url]
            technology = row["value"]
            tech_data = {"technology": technology, "url": parent_url, "host": str(parent_event.host)}
            await self.emit_event(
                tech_data,
                "TECHNOLOGY",
                parent=parent_event,
                context=f"{{module}} visited {parent_url} and found {{event.type}}: {technology}",
            )

    def construct_command(self):
        # base executable
        command = ["gowitness"]
        # chrome path
        if self.chrome_path is not None:
            command += ["--chrome-path", str(self.chrome_path)]
        # db path
        command += ["--db-path", str(self.db_path)]
        # screenshot path
        command += ["--screenshot-path", str(self.screenshot_path)]
        # user agent
        command += ["--user-agent", f"{self.scan.useragent}"]
        # proxy
        if self.proxy:
            command += ["--proxy", str(self.proxy)]
        # resolution
        command += ["--resolution-x", str(self.resolution_x)]
        command += ["--resolution-y", str(self.resolution_y)]
        # input
        command += ["file", "-f", "-"]
        # threads
        command += ["--threads", str(self.threads)]
        # timeout
        command += ["--timeout", str(self.timeout)]
        return command

    async def get_new_screenshots(self):
        screenshots = {}
        if self.db_path.is_file():
            async with aiosqlite.connect(str(self.db_path)) as con:
                con.row_factory = aiosqlite.Row
                con.text_factory = self.helpers.smart_decode
                async with con.execute("SELECT * FROM urls") as cur:
                    async for row in cur:
                        row = dict(row)
                        _id = row["id"]
                        if _id not in self.screenshots_taken:
                            self.screenshots_taken[_id] = row["url"]
                            screenshots[_id] = row
        return screenshots

    async def get_new_network_logs(self):
        network_logs = {}
        if self.db_path.is_file():
            async with aiosqlite.connect(str(self.db_path)) as con:
                con.row_factory = aiosqlite.Row
                async with con.execute("SELECT * FROM network_logs") as cur:
                    async for row in cur:
                        row = dict(row)
                        url = row["final_url"]
                        if url not in self.connections_logged:
                            self.connections_logged.add(url)
                            network_logs[url] = row
        return network_logs

    async def get_new_technologies(self):
        technologies = {}
        if self.db_path.is_file():
            async with aiosqlite.connect(str(self.db_path)) as con:
                con.row_factory = aiosqlite.Row
                async with con.execute("SELECT * FROM technologies") as cur:
                    async for row in cur:
                        _id = row["id"]
                        if _id not in self.technologies_found:
                            self.technologies_found.add(_id)
                            row = dict(row)
                            technologies[_id] = row
        return technologies

    async def cur_execute(self, cur, query):
        try:
            return await cur.execute(query)
        except aiosqlite.OperationalError as e:
            self.warning(f"Error executing query: {query}: {e}")
            return []

    async def report(self):
        if self.screenshots_taken:
            self.success(f"{len(self.screenshots_taken):,} web screenshots captured. To view:")
            self.success("    - Start gowitness")
            self.success(f"        - cd {self.base_path} && ./gowitness server")
            self.success("    - Browse to http://localhost:7171")
        else:
            self.info("No web screenshots captured")



from .base import ModuleTestBase


class TestIP2Location(ModuleTestBase):
    targets = ["8.8.8.8"]
    config_overrides = {"modules": {"ip2location": {"api_key": "asdf"}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="http://api.ip2location.io/?key=asdf&ip=8.8.8.8&format=json&source=bbot",
            json={
                "ip": "8.8.8.8",
                "country_code": "US",
                "country_name": "United States of America",
                "region_name": "California",
                "city_name": "Mountain View",
                "latitude": 37.405992,
                "longitude": -122.078515,
                "zip_code": "94043",
                "time_zone": "-07:00",
                "asn": "15169",
                "as": "Google LLC",
                "is_proxy": False,
            },
        )

    def check(self, module_test, events):
        assert any(
            e.type == "GEOLOCATION" and e.data["ip"] == "8.8.8.8" and e.data["city_name"] == "Mountain View"
            for e in events
        ), "Failed to geolocate IP"



# securitytxt.py
#
# Checks for/parses https://target.domain/.well-known/security.txt
#
# Refer to: https://securitytxt.org/
#
# security.txt may contain email addresses and URL's, and possibly IP addresses.
#
# Example security.txt:
#
#   Contact: mailto:security.reports@example.com
#   Expires: 2028-05-31T14:00:00.000Z
#   Encryption: https://example.com/security.pgp
#   Preferred-Languages: en, es
#   Canonical: https://example.com/.well-known/security.txt
#   Canonical: https://www.example.com/.well-known/security.txt
#   Policy: https://example.com/security-policy.html
#   Hiring: https://example.com/jobs.html
#
# Example security.txt with PGP signature:
#
#   -----BEGIN PGP SIGNED MESSAGE-----
#   Hash: SHA512
#
#   Contact: https://vdp.example.com
#   Expires: 2025-01-01T00:00:00.000Z
#   Preferred-Languages: fr, en
#   Canonical: https://example.com/.well-known/security.txt
#   Policy: https://example.com/cert
#   Hiring: https://www.careers.example.com
#   -----BEGIN PGP SIGNATURE-----
#
#   iQIzBAEBCgAdFiEELC1a63jHPhyV60KPsvWy9dDkrigFAmJBypcACgkQsvWy9dDk
#   rijXHQ//Qya3hUSy5PYW+fI3eFP1+ak6gYq3Cbzkf57cqiBhxGetIGIGNJ6mxgjS
#   KAuvXLMUWgZD73r//fjZ5v1lpuWmpt54+ecat4DgcVCvFKYpaH+KBlay8SX7XtQH
#   9T2NXMcez353TMR3EUOdLwdBzGZprf0Ekg9EzaHKMk0k+A4D9CnSb8Y6BKDPC7wr
#   eadwDIR9ESo0va4sjjcllCG9MF5hqK25SfsKriCSEAMhse2FToEBbw8ImkPKowMN
#   whJ4MIVlBxybu6XoIyk3n7HRRduijywy7uV80pAkhk/hL6wiW3M956FiahfRI6ad
#   +Gky/Ri5TjwAE/x5DhUH8O2toPsn71DeIE4geKfz5d/v41K0yncdrHjzbj0CAHu3
#   wVWLKnEp8RVqTlOR8jU0HqQUQy8iZk4LY91ROv+QjG/jUTWlwun8Ljh+YUeJTMRp
#   MGftCdCrrYjIy5aEQqWztt+dXKac/9e1plq3yyfuW1L+wG3zS7X+NpIJgygMvEwT
#   L3dqfQf63sjk8kWIZMVnicHBlc6BiLqUn020l+pkIOr4MuuJmIlByhlnfqH7YM8k
#   VShwDx7rs4Hj08C7NVCYIySaM2jM4eNKGt9V5k1F1sklCVfYaT8OqOhJrzhcisOC
#   YcQDhjt/iZTR8SzrHO7kFZbaskIp2P7JMaPax2fov15AnNHQQq8=
#   =8vfR
#   -----END PGP SIGNATURE-----

from bbot.modules.base import BaseModule

import re

from bbot.core.helpers.regexes import email_regex, url_regexes

_securitytxt_regex = r"^(?P<k>\w+): *(?P<v>.*)$"
securitytxt_regex = re.compile(_securitytxt_regex, re.I | re.M)


class securitytxt(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["EMAIL_ADDRESS", "URL_UNVERIFIED"]
    flags = ["subdomain-enum", "cloud-enum", "active", "web-basic", "safe"]
    meta = {
        "description": "Check for security.txt content",
        "author": "@colin-stubbs",
        "created_date": "2024-05-26",
    }
    options = {
        "emails": True,
        "urls": True,
    }
    options_desc = {
        "emails": "emit EMAIL_ADDRESS events",
        "urls": "emit URL_UNVERIFIED events",
    }

    async def setup(self):
        self._emails = self.config.get("emails", True)
        self._urls = self.config.get("urls", True)
        return await super().setup()

    def _incoming_dedup_hash(self, event):
        # dedupe by parent
        parent_domain = self.helpers.parent_domain(event.data)
        return hash(parent_domain), "already processed parent domain"

    async def filter_event(self, event):
        if "_wildcard" in str(event.host).split("."):
            return False, "event is wildcard"
        return True

    async def handle_event(self, event):
        tags = ["securitytxt-policy"]
        url = f"https://{event.host}/.well-known/security.txt"

        r = await self.helpers.request(url, method="GET")

        if r is None or r.status_code != 200:
            # it doesn't look like we got a valid response...
            return

        try:
            s = r.text
        except Exception:
            s = ""

        # avoid parsing the response unless it looks, at a very basic level, like an actual security.txt
        s_lower = s.lower()
        if "contact: " in s_lower or "expires: " in s_lower:
            for securitytxt_match in securitytxt_regex.finditer(s):
                v = securitytxt_match.group("v")

                for match in email_regex.finditer(v):
                    start, end = match.span()
                    email = v[start:end]

                    if self._emails:
                        await self.emit_event(email, "EMAIL_ADDRESS", parent=event, tags=tags)

                for url_regex in url_regexes:
                    for match in url_regex.finditer(v):
                        start, end = match.span()
                        found_url = v[start:end]

                        if found_url != url and self._urls is True:
                            await self.emit_event(found_url, "URL_UNVERIFIED", parent=event, tags=tags)


# EOF



import asyncio
import logging
import traceback
from sys import exc_info
from contextlib import suppress

from ..errors import ValidationError
from ..core.helpers.misc import get_size  # noqa
from ..core.helpers.async_helpers import TaskCounter, ShuffleQueue


class BaseModule:
    """The base class for all BBOT modules.

    Attributes:
        watched_events (List): Event types to watch.

        produced_events (List): Event types to produce.

        meta (Dict): Metadata about the module, such as whether authentication is required and a description.

        flags (List): Flags indicating the type of module (must have at least "safe" or "aggressive" and "passive" or "active").

        deps_modules (List): Other BBOT modules this module depends on. Empty list by default.

        deps_pip (List): Python dependencies to install via pip. Empty list by default.

        deps_apt (List): APT package dependencies to install. Empty list by default.

        deps_shell (List): Other dependencies installed via shell commands. Uses [ansible.builtin.shell](https://docs.ansible.com/ansible/latest/collections/ansible/builtin/shell_module.html). Empty list by default.

        deps_ansible (List): Additional Ansible tasks for complex dependencies. Empty list by default.

        accept_dupes (bool): Whether to accept incoming duplicate events. Default is False.

        suppress_dupes (bool): Whether to suppress outgoing duplicate events. Default is True.

        per_host_only (bool): Limit the module to only scanning once per host. Default is False.

        per_hostport_only (bool): Limit the module to only scanning once per host:port. Default is False.

        per_domain_only (bool): Limit the module to only scanning once per domain. Default is False.

        scope_distance_modifier (int, None): Modifies scope distance acceptance for events. Default is 0.
            ```
            None == accept all events
            2 == accept events up to and including the scan's configured search distance plus two
            1 == accept events up to and including the scan's configured search distance plus one
            0 == (DEFAULT) accept events up to and including the scan's configured search distance
            ```

        target_only (bool): Accept only the initial target event(s). Default is False.

        in_scope_only (bool): Accept only explicitly in-scope events, regardless of the scan's search distance. Default is False.

        options (Dict): Customizable options for the module, e.g., {"api_key": ""}. Empty dict by default.

        options_desc (Dict): Descriptions for options, e.g., {"api_key": "API Key"}. Empty dict by default.

        module_threads (int): Maximum concurrent instances of handle_event() or handle_batch(). Default is 1.

        batch_size (int): Size of batches processed by handle_batch(). Default is 1.

        batch_wait (int): Seconds to wait before force-submitting a batch. Default is 10.

        api_failure_abort_threshold (int): Threshold for setting error state after failed HTTP requests (only takes effect when `api_request()` is used. Default is 5.

        _preserve_graph (bool): When set to True, accept events that may be duplicates but are necessary for construction of complete graph. Typically only enabled for output modules that need to maintain full chains of events, e.g. `neo4j` and `json`. Default is False.

        _stats_exclude (bool): Whether to exclude this module from scan statistics. Default is False.

        _qsize (int): Outgoing queue size (0 for infinite). Default is 0.

        _priority (int): Priority level of the module. Lower values are higher priority. Default is 3.

        _name (str): Module name, overridden automatically. Default is 'base'.

        _type (str): Module type, for differentiating between normal and output modules. Default is 'scan'.
    """

    watched_events = []
    produced_events = []
    meta = {"auth_required": False, "description": "Base module"}
    flags = []
    options = {}
    options_desc = {}

    deps_modules = []
    deps_pip = []
    deps_apt = []
    deps_shell = []
    deps_ansible = []

    accept_dupes = False
    suppress_dupes = True
    per_host_only = False
    per_hostport_only = False
    per_domain_only = False
    scope_distance_modifier = 0
    target_only = False
    in_scope_only = False

    _module_threads = 1
    _batch_size = 1
    batch_wait = 10

    # API retries, etc.
    _api_retries = 2
    # disable the module after this many failed attempts in a row
    _api_failure_abort_threshold = 3
    # sleep for this many seconds after being rate limited
    _429_sleep_interval = 30

    default_discovery_context = "{module} discovered {event.type}: {event.data}"

    _preserve_graph = False
    _stats_exclude = False
    _qsize = 1000
    _priority = 3
    _name = "base"
    _type = "scan"
    _intercept = False
    _shuffle_incoming_queue = True

    def __init__(self, scan):
        """Initializes a module instance.

        Args:
            scan: The BBOT scan object associated with this module instance.

        Attributes:
            scan: The scan object associated with this module.

            errored (bool): Whether the module has errored out. Default is False.
        """
        self.scan = scan
        self.errored = False
        self._log = None
        self._incoming_event_queue = None
        self._outgoing_event_queue = None
        # track incoming events to prevent unwanted duplicates
        self._incoming_dup_tracker = set()
        # tracks which subprocesses are running under this module
        self._proc_tracker = set()
        # seconds since we've submitted a batch
        self._last_submitted_batch = None
        # additional callbacks to be executed alongside self.cleanup()
        self.cleanup_callbacks = []
        self._cleanedup = False
        self._watched_events = None

        self._task_counter = TaskCounter()

        # string constant
        self._custom_filter_criteria_msg = "it did not meet custom filter criteria"

        self._api_keys = []

        # track number of failures (for .api_request())
        self._api_request_failures = 0

        self._tasks = []
        self._event_received = None

        # used for optional "per host" tracking
        self._per_host_tracker = set()

    async def setup(self):
        """
        Performs one-time setup tasks for the module.

        This method is responsible for preparing the module for its operation, which may include tasks
        such as downloading necessary resources, validating configuration parameters, or other preliminary
        checks.

        Returns:
            tuple:
                - bool or None: A status indicating the outcome of the setup process. Returns `True` if
                the setup was successful, `None` for a soft-fail where the module setup did not succeed
                but the scan will continue with the module disabled, and `False` for a hard-fail where
                the setup failure causes the scan to abort.
                - str, optional: A reason for the setup failure, provided only when the setup does not
                succeed (i.e., returns `None` or `False`).

        Examples:
            >>> async def setup(self):
            >>>     if not self.config.get("api_key"):
            >>>         # Soft-fail: Configuration missing an API key
            >>>         return None, "No API key specified"

            >>> async def setup(self):
            >>>     try:
            >>>         wordlist = await self.helpers.wordlist("https://raw.githubusercontent.com/user/wordlist.txt")
            >>>     except WordlistError as e:
            >>>         # Hard-fail: Error retrieving wordlist
            >>>         return False, f"Error retrieving wordlist: {e}"

            >>> async def setup(self):
            >>>     self.timeout = self.config.get("timeout", 5)
            >>>     # Success: Setup completed without issues
            >>>     return True
        """

        return True

    async def handle_event(self, event, **kwargs):
        """Asynchronously handles incoming events that the module is configured to watch.

        This method is automatically invoked when an event that matches any in `watched_events` is encountered during a scan. Override this method to implement custom event-handling logic for your module.

        Args:
            event (Event): The event object containing details about the incoming event.

        Note:
            This method should be overridden if the `batch_size` attribute of the module is set to 1.

        Returns:
            None
        """
        pass

    async def handle_batch(self, *events):
        """Handles incoming events in batches for optimized processing.

        This method is automatically called when multiple events that match any in `watched_events` are encountered and the `batch_size` attribute is set to a value greater than 1. Override this method to implement custom batch event-handling logic for your module.

        Args:
            *events (Event): A variable number of Event objects to be processed in a batch.

        Note:
            This method should be overridden if the `batch_size` attribute of the module is set to a value greater than 1.

        Returns:
            None
        """
        pass

    async def filter_event(self, event):
        """Asynchronously filters incoming events based on custom criteria.

        Override this method for more granular control over which events are accepted by your module. This method is called automatically before `handle_event()` for each incoming event that matches any in `watched_events`.

        Args:
            event (Event): The incoming Event object to be filtered.

        Returns:
            tuple: A 2-tuple where the first value is a bool indicating whether the event should be accepted, and the second value is a string explaining the reason for its acceptance or rejection. By default, returns `(True, None)` to indicate acceptance without reason.

        Note:
            This method should be overridden if the module requires custom logic for event filtering.
        """
        return True

    async def finish(self):
        """Asynchronously performs final tasks as the scan nears completion.

        This method can be overridden to execute any necessary finalization logic. For example, if the module relies on a word cloud, you might wait for the scan to finish to ensure the word cloud is most complete before running an operation.

        Returns:
            None

        Warnings:
            This method may be called multiple times since it can raise events, which may re-trigger the "finish" phase of the scan. Optional to override.
        """
        return

    async def report(self):
        """Asynchronously executes a final task after the scan is complete but before cleanup.

        This method can be overridden to aggregate data and raise summary events at the end of the scan.

        Returns:
            None

        Note:
            This method is called only once per scan.
        """
        return

    async def cleanup(self):
        """Asynchronously performs final cleanup operations after the scan is complete.

        This method can be overridden to implement custom cleanup logic. It is called only once per scan and may not raise events.

        Returns:
            None

        Note:
            This method is called only once per scan and may not raise events.
        """
        return

    async def require_api_key(self):
        """
        Asynchronously checks if an API key is required and valid.

        Args:
            None

        Returns:
            bool or tuple: Returns True if API key is valid and ready.
                          Returns a tuple (None, "error message") otherwise.

        Notes:
            - Fetches the API key from the configuration.
            - Calls the 'ping()' method to test API accessibility.
            - Sets the API key readiness status accordingly.
        """
        self.api_key = self.config.get("api_key", "")
        if self.auth_secret:
            try:
                await self.ping()
                self.hugesuccess("API is ready")
                return True, ""
            except Exception as e:
                self.trace(traceback.format_exc())
                return None, f"Error with API ({str(e).strip()})"
        else:
            return None, "No API key set"

    @property
    def api_key(self):
        if self._api_keys:
            return self._api_keys[0]

    @api_key.setter
    def api_key(self, api_keys):
        if isinstance(api_keys, str):
            api_keys = [api_keys]
        self._api_keys = list(api_keys)

    def cycle_api_key(self):
        if len(self._api_keys) > 1:
            self.verbose("Cycling API key")
            self._api_keys.insert(0, self._api_keys.pop())
        else:
            self.debug("No extra API keys to cycle")

    @property
    def api_retries(self):
        return max(self._api_retries + 1, len(self._api_keys))

    @property
    def api_failure_abort_threshold(self):
        return (self.api_retries * self._api_failure_abort_threshold) + 1

    async def ping(self, url=None):
        """Asynchronously checks the health of the configured API.

        This method is used in conjunction with require_api_key() to verify that the API is not just configured, but also responsive. It makes a test request to a known endpoint to validate the API's health.

        The method uses the `ping_url` attribute if defined, or falls back to a provided URL. If neither is available, no request is made.

        Args:
            url (str, optional): A specific URL to use for the ping request. If not provided, the method will use the `ping_url` attribute.

        Returns:
            None

        Raises:
            ValueError: If the API response is not successful (status code != 200).

        Example Usage:
            To use this method, simply define the `ping_url` attribute in your module:

            class MyModule(BaseModule):
                ping_url = "https://api.example.com/ping"

            Alternatively, you can override this method for more complex health checks:

            async def ping(self):
                r = await self.api_request(f"{self.base_url}/complex-health-check")
                if r.status_code != 200 or r.json().get('status') != 'healthy':
                    raise ValueError(f"API unhealthy: {r.text}")
        """
        if url is None:
            url = getattr(self, "ping_url", "")
        if url:
            r = await self.api_request(url)
            if getattr(r, "status_code", 0) != 200:
                response_text = getattr(r, "text", "no response from server")
                raise ValueError(response_text)

    @property
    def batch_size(self):
        batch_size = self.config.get("batch_size", None)
        # only allow overriding the batch size if its default value is greater than 1
        # this prevents modules from being accidentally neutered by an incorrect batch_size setting
        if batch_size is None or self._batch_size == 1:
            batch_size = self._batch_size
        return batch_size

    @property
    def module_threads(self):
        module_threads = self.config.get("module_threads", None)
        if module_threads is None:
            module_threads = self._module_threads
        return module_threads

    @property
    def auth_secret(self):
        """Indicates if the module is properly configured for authentication.

        This read-only property should be used to check whether all necessary attributes (e.g., API keys, tokens, etc.) are configured to perform authenticated requests in the module. Commonly used in setup or initialization steps.

        Returns:
            bool: True if the module is properly configured for authentication, otherwise False.
        """
        return getattr(self, "api_key", "")

    @property
    def event_received(self):
        if self._event_received is None:
            self._event_received = asyncio.Condition()
        return self._event_received

    def get_watched_events(self):
        """Retrieve the set of events that the module is interested in observing.

        Override this method if the set of events the module should watch needs to be determined dynamically, e.g., based on configuration options or other runtime conditions.

        Returns:
            set: The set of event types that this module will handle.
        """
        if self._watched_events is None:
            self._watched_events = set(self.watched_events)
        return self._watched_events

    async def _handle_batch(self):
        """
        Asynchronously handles a batch of events in the module.

        Args:
            None

        Returns:
            bool: True if events were submitted for processing, False otherwise.

        Notes:
            - The method is wrapped in a task counter to monitor asynchronous operations.
            - Checks if there are any events in the incoming queue and module is not in an error state.
            - Invokes '_events_waiting()' to fetch a batch of events.
            - Calls the module's 'handle_batch()' method to process these events.
            - If a "FINISHED" event is found, invokes 'finish()' method of the module.
        """
        finish = False
        async with self._task_counter.count(f"{self.name}.handle_batch()") as counter:
            submitted = False
            if self.batch_size <= 1:
                return
            if self.num_incoming_events > 0:
                events, finish = await self._events_waiting()
                if events and not self.errored:
                    counter.n = len(events)
                    self.verbose(f"Handling batch of {len(events):,} events")
                    submitted = True
                    async with self.scan._acatch(f"{self.name}.handle_batch()"):
                        await self.handle_batch(*events)
                    self.verbose(f"Finished handling batch of {len(events):,} events")
        if finish:
            context = f"{self.name}.finish()"
            async with self.scan._acatch(context), self._task_counter.count(context):
                await self.finish()
        return submitted

    def make_event(self, *args, **kwargs):
        """Create an event for the scan.

        Raises a validation error if the event could not be created, unless raise_error is set to False.

        Args:
            *args: Positional arguments to be passed to the scan's make_event method.
            **kwargs: Keyword arguments to be passed to the scan's make_event method.
            raise_error (bool, optional): Whether to raise a validation error if the event could not be created. Defaults to False.

        Examples:
            >>> new_event = self.make_event("1.2.3.4", parent=event)
            >>> await self.emit_event(new_event)

        Returns:
            Event or None: The created event, or None if a validation error occurred and raise_error was False.

        Raises:
            ValidationError: If the event could not be validated and raise_error is True.
        """
        raise_error = kwargs.pop("raise_error", False)
        module = kwargs.pop("module", None)
        if module is None:
            if (not args) or getattr(args[0], "module", None) is None:
                kwargs["module"] = self
        try:
            event = self.scan.make_event(*args, **kwargs)
        except ValidationError as e:
            if raise_error:
                raise
            self.warning(f"{e}")
            return
        return event

    async def emit_event(self, *args, **kwargs):
        """Emit an event to the event queue and distribute it to interested modules.

        This is how modules "return" data.

        The method first creates an event object by calling `self.make_event()` with the provided arguments.
        Then, the event is queued for outgoing distribution using `self.queue_outgoing_event()`.

        Args:
            *args: Positional arguments to be passed to `self.make_event()` for event creation.
            **kwargs: Keyword arguments to be passed for event creation or configuration of the emit action.
                ```markdown
                - on_success_callback: Optional callback function to execute upon successful event emission.
                - abort_if: Optional condition under which the event emission should be aborted.
                - quick: Optional flag to indicate whether the event should be processed quickly.
                ```

        Examples:
            >>> await self.emit_event("www.evilcorp.com", parent=event, tags=["affiliate"])

            >>> new_event = self.make_event("1.2.3.4", parent=event)
            >>> await self.emit_event(new_event)

        Returns:
            None

        Raises:
            ValidationError: If the event cannot be validated (handled in `self.make_event()`).
        """
        event_kwargs = dict(kwargs)
        emit_kwargs = {}
        for o in ("on_success_callback", "abort_if", "quick"):
            v = event_kwargs.pop(o, None)
            if v is not None:
                emit_kwargs[o] = v
        event = self.make_event(*args, **event_kwargs)
        if event:
            await self.queue_outgoing_event(event, **emit_kwargs)
        return event

    async def _events_waiting(self, batch_size=None):
        """
        Asynchronously fetches events from the incoming_event_queue, up to a specified batch size.

        Args:
            None

        Returns:
            tuple: A tuple containing two elements:
                - events (list): A list of acceptable events from the queue.
                - finish (bool): A flag indicating if a "FINISHED" event is encountered.

        Notes:
            - The method pulls events from incoming_event_queue using 'get_nowait()'.
            - Events go through '_event_postcheck()' for validation.
            - "FINISHED" events are handled differently and the finish flag is set to True.
            - If the queue is empty or the batch size is reached, the loop breaks.
        """
        if batch_size is None:
            batch_size = self.batch_size
        events = []
        finish = False
        while self.incoming_event_queue:
            if batch_size != -1 and len(events) > self.batch_size:
                break
            try:
                event = self.incoming_event_queue.get_nowait()
                self.debug(f"Got {event} from {getattr(event, 'module', 'unknown_module')}")
                acceptable, reason = await self._event_postcheck(event)
                if acceptable:
                    if event.type == "FINISHED":
                        finish = True
                    else:
                        events.append(event)
                        self.scan.stats.event_consumed(event, self)
                elif reason:
                    self.debug(f"Not accepting {event} because {reason}")
            except asyncio.queues.QueueEmpty:
                break
        return events, finish

    @property
    def num_incoming_events(self):
        ret = 0
        if self.incoming_event_queue is not False:
            ret = self.incoming_event_queue.qsize()
        return ret

    def start(self):
        self._tasks = [
            asyncio.create_task(self._worker(), name=f"{self.scan.name}.{self.name}._worker()")
            for _ in range(self.module_threads)
        ]

    async def _setup(self):
        """
        Asynchronously sets up the module by invoking its 'setup()' method.

        This method catches exceptions during setup, sets the module's error state if necessary, and determines the
        status code based on the result of the setup process.

        Args:
            None

        Returns:
            tuple: A tuple containing the module's name, status (True for success, False for hard-fail, None for soft-fail),
            and an optional status message.

        Raises:
            Exception: Captured exceptions from the 'setup()' method are logged, but not propagated.

        Notes:
            - The 'setup()' method can return either a simple boolean status or a tuple of status and message.
            - A WordlistError exception triggers a soft-fail status.
            - The debug log will contain setup status information for the module.
        """
        status_codes = {False: "hard-fail", None: "soft-fail", True: "success"}

        status = False
        self.debug(f"Setting up module {self.name}")
        try:
            result = await self.setup()
            if type(result) == tuple and len(result) == 2:
                status, msg = result
            else:
                status = result
                msg = status_codes[status]
            self.debug(f"Finished setting up module {self.name}")
        except Exception as e:
            self.set_error_state(f"Unexpected error during module setup: {e}", critical=True)
            msg = f"{e}"
            self.trace()
        return self, status, str(msg)

    async def _worker(self):
        """
        The core worker loop for the module, responsible for handling events from the incoming event queue.

        This method is a coroutine and is run asynchronously. Multiple instances can run simultaneously based on
        the 'module_threads' configuration. The worker dequeues events from 'incoming_event_queue', performs
        necessary prechecks, and passes the event to the appropriate handler function.

        Args:
            None

        Returns:
            None

        Raises:
            asyncio.CancelledError: If the worker is cancelled during its operation.

        Notes:
            - The worker is sensitive to the 'stopping' flag of the scan. It will terminate if this flag is set.
            - The worker handles backpressure by pausing when the outgoing event queue is full.
            - Batch processing is supported and is activated when 'batch_size' > 1.
            - Each event is subject to a post-check via '_event_postcheck()' to decide whether it should be handled.
            - Special 'FINISHED' events trigger the 'finish()' method of the module.
        """
        async with self.scan._acatch(context=self._worker, unhandled_is_critical=True):
            try:
                while not self.scan.stopping and not self.errored:
                    # hold the reigns if our outgoing queue is full
                    if self._qsize > 0 and self.outgoing_event_queue.qsize() >= self._qsize:
                        await asyncio.sleep(0.1)
                        continue

                    # if batch wasn't big enough, we wait for the next event before continuing
                    if self.batch_size > 1:
                        submitted = await self._handle_batch()
                        if not submitted:
                            async with self.event_received:
                                await self.event_received.wait()

                    else:
                        try:
                            if self.incoming_event_queue is not False:
                                event = await self.incoming_event_queue.get()
                            else:
                                self.debug("Event queue is in bad state")
                                break
                        except asyncio.queues.QueueEmpty:
                            continue
                        self.debug(f"Got {event} from {getattr(event, 'module', 'unknown_module')}")
                        async with self._task_counter.count(f"event_postcheck({event})"):
                            acceptable, reason = await self._event_postcheck(event)
                        if acceptable:
                            if event.type == "FINISHED":
                                context = f"{self.name}.finish()"
                                async with self.scan._acatch(context), self._task_counter.count(context):
                                    await self.finish()
                            else:
                                context = f"{self.name}.handle_event({event})"
                                self.scan.stats.event_consumed(event, self)
                                self.debug(f"Handling {event}")
                                async with self.scan._acatch(context), self._task_counter.count(context):
                                    await self.handle_event(event)
                                self.debug(f"Finished handling {event}")
                        else:
                            self.debug(f"Not accepting {event} because {reason}")
            except asyncio.CancelledError:
                # this trace was used for debugging leaked CancelledErrors from inside httpx
                # self.log.trace("Worker cancelled")
                raise
            except BaseException as e:
                if self.helpers.in_exception_chain(e, (KeyboardInterrupt,)):
                    self.scan.stop()
                else:
                    self.error(f"Critical failure in module {self.name}: {e}")
                    self.error(traceback.format_exc())
        self.log.trace("Worker stopped")

    @property
    def max_scope_distance(self):
        if self.in_scope_only or self.target_only:
            return 0
        if self.scope_distance_modifier is None:
            return 999
        return max(0, self.scan.scope_search_distance + self.scope_distance_modifier)

    def _event_precheck(self, event):
        """
        Pre-checks an event to determine if it should be accepted by the module for queuing.

        This method is called when an event is about to be enqueued into the module's incoming event queue.
        It applies various filters such as special signal event types, module error state, watched event types, and more
        to decide whether or not the event should be enqueued.

        Args:
            event (Event): The event object to check.

        Returns:
            tuple: A tuple (bool, str) where the bool indicates if the event should be accepted, and the str gives the reason.

        Examples:
            >>> result, reason = self._event_precheck(event)
            >>> if result:
            ...     self.incoming_event_queue.put_nowait(event)
            ... else:
            ...     self.debug(f"Not accepting {event} because {reason}")

        Notes:
            - The method considers special signal event types like "FINISHED".
            - Checks whether the module is in an error state.
            - Checks if the event type matches the types this module is interested in (`watched_events`).
            - Checks for events tagged as 'target' if the module has `target_only` flag set.
            - Applies specific filtering based on event type and module name.
        """

        # special signal event types
        if event.type in ("FINISHED",):
            return True, "its type is FINISHED"
        if self.errored:
            return False, "module is in error state"
        # exclude non-watched types
        if not any(t in self.get_watched_events() for t in ("*", event.type)):
            return False, "its type is not in watched_events"
        if self.target_only:
            if "target" not in event.tags:
                return False, "it did not meet target_only filter criteria"

        # exclude certain URLs (e.g. javascript):
        # TODO: revisit this after httpx rework
        if event.type.startswith("URL") and self.name != "httpx" and "httpx-only" in event.tags:
            return False, "its extension was listed in url_extension_httpx_only"

        return True, "precheck succeeded"

    async def _event_postcheck(self, event):
        """
        A simple wrapper for dup tracking
        """
        # special exception for "FINISHED" event
        if event.type in ("FINISHED",):
            return True, ""
        acceptable, reason = await self._event_postcheck_inner(event)
        if acceptable:
            # check duplicates
            is_incoming_duplicate, reason = self.is_incoming_duplicate(event, add=True)
            if is_incoming_duplicate and not self.accept_dupes:
                return False, "module has already seen it" + (f" ({reason})" if reason else "")

        return acceptable, reason

    async def _event_postcheck_inner(self, event):
        """
        Post-checks an event to determine if it should be accepted by the module for handling.

        This method is called when an event is dequeued from the module's incoming event queue, right before it is actually processed.
        It applies various filters such as scope, custom filtering logic, and per-host tracking to decide the event's fate.

        Args:
            event (Event): The event object to check.

        Returns:
            tuple: A tuple (bool, str) where the bool indicates if the event should be accepted, and the str gives the reason.

        Notes:
            - Override the `filter_event` method for custom filtering logic.
            - This method also maintains host-based tracking when the `per_host_only` or similar flags are set.
            - The method will also update event production stats for output modules.
        """
        # force-output certain events to the graph
        if self._is_graph_important(event):
            return True, "event is critical to the graph"

        # check scope distance
        filter_result, reason = self._scope_distance_check(event)
        if not filter_result:
            return filter_result, reason

        # custom filtering
        async with self.scan._acatch(context=self.filter_event):
            try:
                filter_result = await self.filter_event(event)
            except Exception as e:
                msg = f"Unhandled exception in {self.name}.filter_event({event}): {e}"
                self.error(msg)
                return False, msg
            msg = str(self._custom_filter_criteria_msg)
            with suppress(ValueError, TypeError):
                filter_result, reason = filter_result
                msg += f": {reason}"
            if not filter_result:
                return False, msg

        self.debug(f"{event} passed post-check")
        return True, ""

    def _scope_distance_check(self, event):
        if self.in_scope_only:
            if event.scope_distance > 0:
                return False, "it did not meet in_scope_only filter criteria"
        if self.scope_distance_modifier is not None:
            if event.scope_distance < 0:
                return False, f"its scope_distance ({event.scope_distance}) is invalid."
            elif event.scope_distance > self.max_scope_distance:
                return (
                    False,
                    f"its scope_distance ({event.scope_distance}) exceeds the maximum allowed by the scan ({self.scan.scope_search_distance}) + the module ({self.scope_distance_modifier}) == {self.max_scope_distance}",
                )
        return True, ""

    async def _cleanup(self):
        if not self._cleanedup:
            self._cleanedup = True
            for callback in [self.cleanup] + self.cleanup_callbacks:
                context = f"{self.name}.cleanup()"
                if callable(callback):
                    async with self.scan._acatch(context), self._task_counter.count(context):
                        await self.helpers.execute_sync_or_async(callback)

    async def queue_event(self, event):
        """
        Asynchronously queues an incoming event to the module's event queue for further processing.

        The function performs an initial check to see if the event is acceptable for queuing.
        If the event passes the check, it is put into the `incoming_event_queue`.

        Args:
            event: The event object to be queued.

        Returns:
            None: The function doesn't return anything but modifies the state of the `incoming_event_queue`.

        Examples:
            >>> await self.queue_event(some_event)

        Raises:
            AttributeError: If the module is not in an acceptable state to queue incoming events.
        """
        async with self._task_counter.count("queue_event()", _log=False):
            if self.incoming_event_queue is False:
                self.debug("Not in an acceptable state to queue incoming event")
                return
            acceptable, reason = self._event_precheck(event)
            if not acceptable:
                if reason and reason != "its type is not in watched_events":
                    self.debug(f"Not queueing {event} because {reason}")
                return
            else:
                self.debug(f"Queueing {event} because {reason}")
            try:
                self.incoming_event_queue.put_nowait(event)
                async with self.event_received:
                    self.event_received.notify()
                if event.type != "FINISHED":
                    self.scan._new_activity = True
            except AttributeError:
                self.debug("Not in an acceptable state to queue incoming event")

    async def queue_outgoing_event(self, event, **kwargs):
        """
        Queues an outgoing event to the module's outgoing event queue for further processing.

        The function attempts to put the event into the `outgoing_event_queue` immediately.
        If it's not possible due to the current state of the module, an AttributeError is raised, and a debug log is generated.

        Args:
            event: The event object to be queued.
            **kwargs: Additional keyword arguments to be associated with the event.

        Returns:
            None: The function doesn't return anything but modifies the state of the `outgoing_event_queue`.

        Examples:
            >>> self.queue_outgoing_event(some_outgoing_event, abort_if=lambda e: "unresolved" in e.tags)

        Raises:
            AttributeError: If the module is not in an acceptable state to queue outgoing events.
        """
        try:
            await self.outgoing_event_queue.put((event, kwargs))
        except AttributeError:
            self.debug("Not in an acceptable state to queue outgoing event")

    def set_error_state(self, message=None, clear_outgoing_queue=False, critical=False):
        """
        Puts the module into an errored state where it cannot accept new events. Optionally logs a warning message.

        The function sets the module's `errored` attribute to True and logs a warning with the optional message.
        It also clears the incoming event queue to prevent further processing and updates its status to False.

        Args:
            message (str, optional): Additional message to be logged along with the warning.

        Returns:
            None: The function doesn't return anything but updates the `errored` state and clears the incoming event queue.

        Examples:
            >>> self.set_error_state()
            >>> self.set_error_state("Failed to connect to the server")

        Notes:
            - The function sets `self._incoming_event_queue` to False to prevent its further use.
            - If the module was already in an errored state, the function will not reset the error state or the queue.
        """
        if not self.errored:
            log_msg = "Setting error state"
            if message is not None:
                log_msg += f": {message}"
            if critical:
                log_fn = self.error
            else:
                log_fn = self.warning
            log_fn(log_msg)
            self.errored = True
            # clear incoming queue
            if self.incoming_event_queue is not False:
                self.debug("Emptying event_queue")
                with suppress(asyncio.queues.QueueEmpty):
                    while 1:
                        self.incoming_event_queue.get_nowait()
                # set queue to None to prevent its use
                # if there are leftover objects in the queue, the scan will hang.
                self._incoming_event_queue = False

            if clear_outgoing_queue:
                with suppress(asyncio.queues.QueueEmpty):
                    while 1:
                        self.outgoing_event_queue.get_nowait()

    def is_incoming_duplicate(self, event, add=False):
        if event.type in ("FINISHED",):
            return False, ""
        reason = ""
        try:
            event_hash = self._incoming_dedup_hash(event)
        except Exception as e:
            msg = f"Unhandled exception in {self.name}._incoming_dedup_hash({event}): {e}"
            self.error(msg)
            return True, msg
        with suppress(TypeError, ValueError):
            event_hash, reason = event_hash
        is_dup = event_hash in self._incoming_dup_tracker
        if add:
            self._incoming_dup_tracker.add(event_hash)
        return is_dup, reason

    def _incoming_dedup_hash(self, event):
        """
        Determines the criteria for what is considered to be a duplicate event if `accept_dupes` is False.
        """
        if self.per_host_only:
            return self.get_per_host_hash(event), "per_host_only=True"
        if self.per_hostport_only:
            return self.get_per_hostport_hash(event), "per_hostport_only=True"
        elif self.per_domain_only:
            return self.get_per_domain_hash(event), "per_domain_only=True"
        return hash(event), ""

    def _outgoing_dedup_hash(self, event):
        """
        Determines the criteria for what is considered to be a duplicate event if `suppress_dupes` is True.

        We take into account the `internal` attribute we don't want an internal event (which isn't distributed to output modules)
        to inadvertently suppress a non-internal event.
        """
        return hash((event, self.name, event.internal, event.always_emit))

    def get_per_host_hash(self, event):
        """
        Computes a per-host hash value for a given event. This method may be optionally overridden in subclasses.

        The function uses the event's `host` to create a string to be hashed.

        Args:
            event (Event): The event object containing host information.

        Returns:
            int: The hash value computed for the host.

        Examples:
            >>> event = self.make_event("https://example.com:8443")
            >>> self.get_per_host_hash(event)
        """
        return hash(event.host)

    def get_per_hostport_hash(self, event):
        """
        Computes a per-host:port hash value for a given event. This method may be optionally overridden in subclasses.

        The function uses the event's `host`, `port`, and `scheme` (for URLs) to create a string to be hashed.
        The hash value is used for distinguishing events related to the same host.

        Args:
            event (Event): The event object containing host, port, or parsed URL information.

        Returns:
            int: The hash value computed for the host.

        Examples:
            >>> event = self.make_event("https://example.com:8443")
            >>> self.get_per_hostport_hash(event)
        """
        parsed = getattr(event, "parsed_url", None)
        if parsed is None:
            to_hash = self.helpers.make_netloc(event.host, event.port)
        else:
            to_hash = f"{parsed.scheme}://{parsed.netloc}/"
        return hash(to_hash)

    def get_per_domain_hash(self, event):
        """
        Computes a per-domain hash value for a given event. This method may be optionally overridden in subclasses.

        Events with the same root domain will receive the same hash value.

        Args:
            event (Event): The event object containing host, port, or parsed URL information.

        Returns:
            int: The hash value computed for the domain.

        Examples:
            >>> event = self.make_event("https://www.example.com:8443")
            >>> self.get_per_domain_hash(event)
        """
        _, domain = self.helpers.split_domain(event.host)
        return hash(domain)

    @property
    def name(self):
        return str(self._name)

    @property
    def helpers(self):
        return self.scan.helpers

    @property
    def status(self):
        """
        Provides the current status of the module as a dictionary.

        The dictionary contains the following keys:
            - 'events': A sub-dictionary with 'incoming' and 'outgoing' keys, representing the number of events in the respective queues.
            - 'tasks': The current value of the task counter.
            - 'errored': A boolean value indicating if the module is in an error state.
            - 'running': A boolean value indicating if the module is currently processing data.

        Returns:
            dict: A dictionary containing the current status of the module.

        Examples:
            >>> self.status
            {'events': {'incoming': 5, 'outgoing': 2}, 'tasks': 3, 'errored': False, 'running': True}
        """
        status = {
            "events": {"incoming": self.num_incoming_events, "outgoing": self.outgoing_event_queue.qsize()},
            "tasks": self._task_counter.value,
            "errored": self.errored,
        }
        status["running"] = self.running
        return status

    @property
    def running(self):
        """Property indicating whether the module is currently processing data.

        This property checks if the task counter (`self._task_counter.value`) is greater than zero,
        indicating that there are ongoing tasks in the module.

        Returns:
            bool: True if the module is currently processing data, False otherwise.
        """
        return self._task_counter.value > 0

    @property
    def finished(self):
        """Property indicating whether the module has finished processing.

        This property checks three conditions to determine if the module is finished:
        1. The module is not currently running (`self.running` is False).
        2. The number of incoming events in the queue is zero or less (`self.num_incoming_events <= 0`).
        3. The number of outgoing events in the queue is zero or less (`self.outgoing_event_queue.qsize() <= 0`).

        Returns:
            bool: True if the module has finished processing, False otherwise.
        """
        return not self.running and self.num_incoming_events <= 0 and self.outgoing_event_queue.qsize() <= 0

    async def run_process(self, *args, **kwargs):
        kwargs["_proc_tracker"] = self._proc_tracker
        return await self.helpers.run(*args, **kwargs)

    async def run_process_live(self, *args, **kwargs):
        kwargs["_proc_tracker"] = self._proc_tracker
        async for line in self.helpers.run_live(*args, **kwargs):
            yield line

    def prepare_api_request(self, url, kwargs):
        """
        Prepare an API request by adding the necessary authentication - header, bearer token, etc.
        """
        if self.api_key:
            url = url.format(api_key=self.api_key)
            if "headers" not in kwargs:
                kwargs["headers"] = {}
            kwargs["headers"]["Authorization"] = f"Bearer {self.api_key}"
        return url, kwargs

    async def api_request(self, *args, **kwargs):
        """
        Makes an HTTP request while automatically:
            - avoiding rate limits (sleep/retry)
            - cycling API keys
            - cancelling after too many failed attempts
        """
        url = args[0] if args else kwargs.pop("url", "")

        # loop until we have a successful request
        for _ in range(self.api_retries):
            if "headers" not in kwargs:
                kwargs["headers"] = {}
            new_url, kwargs = self.prepare_api_request(url, kwargs)
            kwargs["url"] = new_url

            r = await self.helpers.request(**kwargs)
            success = r is not None and self._api_response_is_success(r)

            if success:
                self._api_request_failures = 0
            else:
                status_code = getattr(r, "status_code", 0)
                response_text = getattr(r, "text", "")
                self.trace(f"API response to {url} failed with status code {status_code}: {response_text}")
                self._api_request_failures += 1
                if self._api_request_failures >= self.api_failure_abort_threshold:
                    self.set_error_state(
                        f"Setting error state due to {self._api_request_failures:,} failed HTTP requests"
                    )
                else:
                    # sleep for a bit if we're being rate limited
                    retry_after = self._get_retry_after(r)
                    if retry_after or status_code == 429:
                        sleep_interval = int(retry_after) if retry_after is not None else self._429_sleep_interval
                        self.verbose(
                            f"Sleeping for {sleep_interval:,} seconds due to rate limit (HTTP status: {status_code})"
                        )
                        await asyncio.sleep(sleep_interval)
                    elif self._api_keys:
                        # if request failed, cycle API keys and try again
                        self.cycle_api_key()
                    continue
            break

        return r

    def _get_retry_after(self, r):
        # try to get retry_after from headers first
        headers = getattr(r, "headers", {})
        retry_after = headers.get("Retry-After", None)
        if retry_after is None:
            # then look in body json
            with suppress(Exception):
                body_json = r.json()
                if isinstance(body_json, dict):
                    retry_after = body_json.get("retry_after", None)
        if retry_after is not None:
            return float(retry_after)

    def _prepare_api_iter_req(self, url, page, page_size, offset, **requests_kwargs):
        """
        Default function for preparing an API request for iterating through paginated data.
        """
        url = self.helpers.safe_format(url, page=page, page_size=page_size, offset=offset)
        return url, requests_kwargs

    def _api_response_is_success(self, r):
        return r.is_success

    async def api_page_iter(self, url, page_size=100, _json=True, next_key=None, iter_key=None, **requests_kwargs):
        """
        An asynchronous generator function for iterating through paginated API data.

        This function continuously makes requests to a specified API URL, incrementing the page number
        or applying a custom pagination function, and yields the received data one page at a time.
        It is well-suited for APIs that provide paginated results.

        Args:
            url (str): The initial API URL. Can contain placeholders for 'page', 'page_size', and 'offset'.
            page_size (int, optional): The number of items per page. Defaults to 100.
            json (bool, optional): If True, attempts to deserialize the response content to a JSON object. Defaults to True.
            next_key (callable, optional): A function that takes the last page's data and returns the URL for the next page. Defaults to None.
            iter_key (callable, optional): A function that builds each new request based on the page number, page size, and offset. Defaults to a simple implementation that autoreplaces {page} and {page_size} in the url.
            **requests_kwargs: Arbitrary keyword arguments that will be forwarded to the HTTP request function.

        Yields:
            dict or httpx.Response: If 'json' is True, yields a dictionary containing the parsed JSON data. Otherwise, yields the raw HTTP response.

        Note:
            The loop will continue indefinitely unless manually stopped. Make sure to break out of the loop once the last page has been received.

        Examples:
            >>> agen = api_page_iter('https://api.example.com/data?page={page}&page_size={page_size}')
            >>> try:
            >>>     async for page in agen:
            >>>         subdomains = page["subdomains"]
            >>>         self.hugesuccess(subdomains)
            >>>         if not subdomains:
            >>>             break
            >>> finally:
            >>>     await agen.aclose()
        """
        page = 1
        offset = 0
        result = None
        if iter_key is None:
            iter_key = self._prepare_api_iter_req
        while 1:
            if result and callable(next_key):
                try:
                    new_url = next_key(result)
                except Exception as e:
                    self.debug(f"Failed to extract next page of results from {url}: {e}")
                    self.debug(traceback.format_exc())
            else:
                new_url, new_kwargs = iter_key(url, page, page_size, offset, **requests_kwargs)
            result = await self.api_request(new_url, **new_kwargs)
            if result is None:
                self.verbose(f"api_page_iter() got no response for {url}")
                break
            try:
                if _json:
                    result = result.json()
                yield result
            except Exception:
                self.warning(f'Error in api_page_iter() for url: "{new_url}"')
                self.trace(traceback.format_exc())
                break
            finally:
                offset += page_size
                page += 1

    @property
    def preset(self):
        return self.scan.preset

    @property
    def config(self):
        """Property that provides easy access to the module's configuration in the scan's config.

        This property serves as a shortcut to retrieve the module-specific configuration from
        `self.scan.config`. If no configuration is found for this module, an empty dictionary is returned.

        Returns:
            dict: The configuration dictionary specific to this module.
        """
        config = self.scan.config.get("modules", {}).get(self.name, {})
        if config is None:
            config = {}
        return config

    @property
    def incoming_event_queue(self):
        if self._incoming_event_queue is None:
            if self._shuffle_incoming_queue:
                self._incoming_event_queue = ShuffleQueue()
            else:
                self._incoming_event_queue = asyncio.Queue()
        return self._incoming_event_queue

    @property
    def outgoing_event_queue(self):
        if self._outgoing_event_queue is None:
            self._outgoing_event_queue = ShuffleQueue(self._qsize)
        return self._outgoing_event_queue

    @property
    def priority(self):
        """
        Gets the priority level of the module as an integer.

        The priority level is constrained to be between 1 and 5, inclusive.
        A lower value indicates a higher priority.

        Returns:
            int: The priority level of the module, constrained between 1 and 5.

        Examples:
            >>> self.priority
            3
        """
        return int(max(1, min(5, self._priority)))

    @property
    def auth_required(self):
        return self.meta.get("auth_required", False)

    @property
    def http_timeout(self):
        """
        Convenience shortcut to `http_timeout` in the config
        """
        return self.scan.web_config.get("http_timeout", 10)

    @property
    def log(self):
        if getattr(self, "_log", None) is None:
            self._log = logging.getLogger(f"bbot.modules.{self.name}")
        return self._log

    @property
    def memory_usage(self):
        """Property that calculates the current memory usage of the module in bytes.

        This property uses the `get_size` function to estimate the memory consumption
        of the module object. The depth of the object graph traversal is limited to 3 levels
        to avoid performance issues. Commonly shared objects like `self.scan`, `self.helpers`,
        are excluded from the calculation to prevent double-counting.

        Returns:
            int: The estimated memory usage of the module in bytes.
        """
        seen = {self.scan, self.helpers, self.log}  # noqa
        return get_size(self, max_depth=3, seen=seen)

    def __str__(self):
        return self.name

    def log_table(self, *args, **kwargs):
        """Logs a table to the console and optionally writes it to a file.

        This function generates a table using `self.helpers.make_table`, then logs each line
        of the table as an info-level log. If a table_name is provided, it also writes the table to a file.

        Args:
            *args: Variable length argument list to be passed to `self.helpers.make_table`.
            **kwargs: Arbitrary keyword arguments. If 'table_name' is specified, the table will be written to a file.

        Returns:
            str: The generated table as a string.

        Examples:
            >>> self.log_table(['Header1', 'Header2'], [['row1col1', 'row1col2'], ['row2col1', 'row2col2']], table_name="my_table")
        """
        table_name = kwargs.pop("table_name", None)
        max_log_entries = kwargs.pop("max_log_entries", None)
        table = self.helpers.make_table(*args, **kwargs)
        lines_logged = 0
        for line in table.splitlines():
            if max_log_entries is not None and lines_logged > max_log_entries:
                break
            self.info(line)
            lines_logged += 1
        if table_name is not None:
            date = self.helpers.make_date()
            filename = self.scan.home / f"{self.helpers.tagify(table_name)}-table-{date}.txt"
            with open(filename, "w") as f:
                f.write(table)
            self.verbose(f"Wrote {table_name} to {filename}")
        return table

    def _is_graph_important(self, event):
        return self.preserve_graph and getattr(event, "_graph_important", False) and not getattr(event, "_omit", False)

    @property
    def preserve_graph(self):
        preserve_graph = self.config.get("preserve_graph", None)
        if preserve_graph is None:
            preserve_graph = self._preserve_graph
        return preserve_graph

    def debug(self, *args, trace=False, **kwargs):
        """Logs debug messages and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.debug("This is a debug message")
            >>> self.debug("This is a debug message with a trace", trace=True)
        """
        self.log.debug(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def verbose(self, *args, trace=False, **kwargs):
        """Logs messages and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.verbose("This is a verbose message")
            >>> self.verbose("This is a verbose message with a trace", trace=True)
        """
        self.log.verbose(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def hugeverbose(self, *args, trace=False, **kwargs):
        """Logs a whole message in emboldened white text, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.hugeverbose("This is a huge verbose message")
            >>> self.hugeverbose("This is a huge verbose message with a trace", trace=True)
        """
        self.log.hugeverbose(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def info(self, *args, trace=False, **kwargs):
        """Logs informational messages and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.info("This is an informational message")
            >>> self.info("This is an informational message with a trace", trace=True)
        """
        self.log.info(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def hugeinfo(self, *args, trace=False, **kwargs):
        """Logs a whole message in emboldened blue text, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.hugeinfo("This is a huge informational message")
            >>> self.hugeinfo("This is a huge informational message with a trace", trace=True)
        """
        self.log.hugeinfo(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def success(self, *args, trace=False, **kwargs):
        """Logs a success message, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.success("Operation completed successfully")
            >>> self.success("Operation completed with a trace", trace=True)
        """
        self.log.success(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def hugesuccess(self, *args, trace=False, **kwargs):
        """Logs a whole message in emboldened green text, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to False.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.hugesuccess("This is a huge success message")
            >>> self.hugesuccess("This is a huge success message with a trace", trace=True)
        """
        self.log.hugesuccess(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def warning(self, *args, trace=True, **kwargs):
        """Logs a warning message, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to True.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.warning("This is a warning message")
            >>> self.warning("This is a warning message with a trace", trace=False)
        """
        self.log.warning(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def hugewarning(self, *args, trace=True, **kwargs):
        """Logs a whole message in emboldened orange text, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to True.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.hugewarning("This is a huge warning message")
            >>> self.hugewarning("This is a huge warning message with a trace", trace=False)
        """
        self.log.hugewarning(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def error(self, *args, trace=True, **kwargs):
        """Logs an error message, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to True.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.error("This is an error message")
            >>> self.error("This is an error message with a trace", trace=False)
        """
        self.log.error(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    def trace(self, msg=None):
        """Logs the stack trace of the most recently caught exception.

        This method captures the type, value, and traceback of the most recent exception and logs it using the trace level. It is typically used for debugging purposes.

        Anything logged using this method will always be written to the scan's `debug.log`, even if debugging is not enabled.

        Examples:
            >>> try:
            >>>     1 / 0
            >>> except ZeroDivisionError:
            >>>     self.trace()
        """
        if msg is None:
            e_type, e_val, e_traceback = exc_info()
            if e_type is not None:
                self.log.trace(traceback.format_exc())
        else:
            self.log.trace(msg)

    def critical(self, *args, trace=True, **kwargs):
        """Logs a whole message in emboldened red text, and optionally the stack trace of the most recent exception.

        Args:
            *args: Variable-length argument list to pass to the logger.
            trace (bool, optional): Whether to log the stack trace of the most recently caught exception. Defaults to True.
            **kwargs: Arbitrary keyword arguments to pass to the logger.

        Examples:
            >>> self.critical("This is a critical message")
            >>> self.critical("This is a critical message with a trace", trace=False)
        """
        self.log.critical(*args, extra={"scan_id": self.scan.id}, **kwargs)
        if trace:
            self.trace()

    @classmethod
    def help_text(self):
        """
        Returns a string containing help text for the module.
        This includes the module's description, metadata, events, flags, and available options.
        """
        # Retrieve the module's metadata, options, events, and flags
        meta = getattr(self, "meta", {})
        options = getattr(self, "options", {})
        options_desc = getattr(self, "options_desc", {})
        watched_events = getattr(self, "watched_events", [])
        produced_events = getattr(self, "produced_events", [])
        flags = getattr(self, "flags", [])

        help_text = "\n" + "=" * 40 + "\n"
        help_text += f"Module Help: {self.__name__}\n"
        help_text += "=" * 40 + "\n\n"

        for key, value in meta.items():
            help_text += f"{key.replace('_', ' ').title()}: {value}\n"

        help_text += "\nWatched Events:\n"
        help_text += "  " + ", ".join(watched_events) + "\n" if watched_events else "  None\n"

        help_text += "\nProduced Events:\n"
        help_text += "  " + ", ".join(produced_events) + "\n" if produced_events else "  None\n"

        help_text += "\nFlags:\n"
        help_text += "  " + ", ".join(flags) + "\n" if flags else "  None\n"

        help_text += "\nOptions:\n"
        if options:
            for option, default_value in options.items():
                option_description = options_desc.get(option, "No description available.")
                help_text += f"  - {option}:\n"
                help_text += f"      Description: {option_description}\n"
                help_text += f"      Default: {default_value}\n"
        else:
            help_text += "  No options available."

        return help_text


class BaseInterceptModule(BaseModule):
    """
    An Intercept Module is a special type of high-priority module that gets early access to events.

    If you want your module to tag or modify an event before it's distributed to the scan, it should
    probably be an intercept module.

    Examples of intercept modules include `dns` (for DNS resolution and wildcard detection)
    and `cloud` (for detection and tagging of cloud assets).
    """

    accept_dupes = True
    _intercept = True

    async def _worker(self):
        async with self.scan._acatch(context=self._worker, unhandled_is_critical=True):
            try:
                while not self.scan.stopping and not self.errored:
                    try:
                        if self.incoming_event_queue is not False:
                            incoming = await self.get_incoming_event()
                            try:
                                event, kwargs = incoming
                            except ValueError:
                                event = incoming
                                kwargs = {}
                        else:
                            self.debug("Event queue is in bad state")
                            break
                    except asyncio.queues.QueueEmpty:
                        await asyncio.sleep(0.1)
                        continue

                    if event.type == "FINISHED":
                        context = f"{self.name}.finish()"
                        async with self.scan._acatch(context), self._task_counter.count(context):
                            await self.finish()
                        continue

                    acceptable = True
                    async with self._task_counter.count(f"event_precheck({event})"):
                        precheck_pass, reason = self._event_precheck(event)
                    if not precheck_pass:
                        self.debug(f"Not intercepting {event} because precheck failed ({reason})")
                        acceptable = False
                    async with self._task_counter.count(f"event_postcheck({event})"):
                        postcheck_pass, reason = await self._event_postcheck(event)
                    if not postcheck_pass:
                        self.debug(f"Not intercepting {event} because postcheck failed ({reason})")
                        acceptable = False

                    # whether to pass the event on to the rest of the scan
                    # defaults to true, unless handle_event returns False
                    forward_event = True
                    forward_event_reason = ""

                    if acceptable:
                        context = f"{self.name}.handle_event({event, kwargs})"
                        self.scan.stats.event_consumed(event, self)
                        self.debug(f"Intercepting {event}")
                        async with self.scan._acatch(context), self._task_counter.count(context):
                            forward_event = await self.handle_event(event, **kwargs)
                            with suppress(ValueError, TypeError):
                                forward_event, forward_event_reason = forward_event

                        if forward_event is False:
                            self.debug(f"Not forwarding {event} because {forward_event_reason}")
                            continue

                    self.debug(f"Forwarding {event}")
                    await self.forward_event(event, kwargs)

            except asyncio.CancelledError:
                # this trace was used for debugging leaked CancelledErrors from inside httpx
                # self.log.trace("Worker cancelled")
                raise
            except BaseException as e:
                if self.helpers.in_exception_chain(e, (KeyboardInterrupt,)):
                    self.scan.stop()
                else:
                    self.critical(f"Critical failure in intercept module {self.name}: {e}")
                    self.critical(traceback.format_exc())
        self.log.trace("Worker stopped")

    async def get_incoming_event(self):
        """
        Get an event from this module's incoming event queue
        """
        return await self.incoming_event_queue.get()

    async def forward_event(self, event, kwargs):
        """
        Used for forwarding the event on to the next intercept module
        """
        await self.outgoing_event_queue.put((event, kwargs))

    async def queue_outgoing_event(self, event, **kwargs):
        """
        Used by emit_event() to raise new events to the scan
        """
        # if this was a normal module, we'd put it in the outgoing queue
        # but because it's an intercept module, we need to queue it at the scan's ingress
        await self.scan.ingress_module.queue_event(event, kwargs)

    async def queue_event(self, event, kwargs=None):
        """
        Put an event in this module's incoming event queue
        """
        if kwargs is None:
            kwargs = {}
        try:
            self.incoming_event_queue.put_nowait((event, kwargs))
        except AttributeError:
            self.debug("Not in an acceptable state to queue incoming event")

    async def _event_postcheck(self, event):
        return await self._event_postcheck_inner(event)



from bbot.modules.base import BaseInterceptModule


class portfilter(BaseInterceptModule):
    watched_events = ["OPEN_TCP_PORT", "URL_UNVERIFIED", "URL"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Filter out unwanted open ports from cloud/CDN targets",
        "created_date": "2025-01-06",
        "author": "@TheTechromancer",
    }
    options = {
        "cdn_tags": "cdn-",
        "allowed_cdn_ports": "80,443",
    }
    options_desc = {
        "cdn_tags": "Comma-separated list of tags to skip, e.g. 'cdn,cloud'",
        "allowed_cdn_ports": "Comma-separated list of ports that are allowed to be scanned for CDNs",
    }

    _priority = 4

    async def setup(self):
        self.cdn_tags = [t.strip() for t in self.config.get("cdn_tags", "").split(",")]
        self.allowed_cdn_ports = self.config.get("allowed_cdn_ports", "").strip()
        if self.allowed_cdn_ports:
            try:
                self.allowed_cdn_ports = [int(p.strip()) for p in self.allowed_cdn_ports.split(",")]
            except Exception as e:
                return False, f"Error parsing allowed CDN ports '{self.allowed_cdn_ports}': {e}"
        return True

    async def handle_event(self, event, **kwargs):
        # if the port isn't in our list of allowed CDN ports
        if event.port not in self.allowed_cdn_ports:
            for cdn_tag in self.cdn_tags:
                # and if any of the event's tags match our CDN filter
                if any(t.startswith(str(cdn_tag)) for t in event.tags):
                    return (
                        False,
                        f"one of the event's tags matches the tag '{cdn_tag}' and the port is not in the allowed list",
                    )
        return True



import random
import ipaddress

from bbot.core.helpers import validators
from bbot.modules.internal.base import BaseInternalModule


class speculate(BaseInternalModule):
    """
    Bridge the gap between ranges and ips, or ips and open ports
    in situations where e.g. a port scanner isn't enabled
    """

    watched_events = [
        "IP_RANGE",
        "URL",
        "URL_UNVERIFIED",
        "DNS_NAME",
        "DNS_NAME_UNRESOLVED",
        "IP_ADDRESS",
        "HTTP_RESPONSE",
        "STORAGE_BUCKET",
        "SOCIAL",
        "AZURE_TENANT",
        "USERNAME",
    ]
    produced_events = ["DNS_NAME", "OPEN_TCP_PORT", "IP_ADDRESS", "FINDING", "ORG_STUB"]
    flags = ["passive"]
    meta = {
        "description": "Derive certain event types from others by common sense",
        "created_date": "2022-05-03",
        "author": "@liquidsec",
    }

    options = {"max_hosts": 65536, "ports": "80,443", "essential_only": False}
    options_desc = {
        "max_hosts": "Max number of IP_RANGE hosts to convert into IP_ADDRESS events",
        "ports": "The set of ports to speculate on",
        "essential_only": "Only enable essential speculate features (no extra discovery)",
    }
    scope_distance_modifier = 1
    _priority = 4

    default_discovery_context = "speculated {event.type}: {event.data}"

    async def setup(self):
        scan_modules = [m for m in self.scan.modules.values() if m._type == "scan"]
        self.open_port_consumers = any("OPEN_TCP_PORT" in m.watched_events for m in scan_modules)
        # only consider active portscanners (still speculate if only passive ones are enabled)
        self.portscanner_enabled = any(
            "portscan" in m.flags and "active" in m.flags for m in self.scan.modules.values()
        )
        self.emit_open_ports = self.open_port_consumers and not self.portscanner_enabled
        self.range_to_ip = True
        self.dns_disable = self.scan.config.get("dns", {}).get("disable", False)
        self.essential_only = self.config.get("essential_only", False)
        self.org_stubs_seen = set()

        port_string = self.config.get("ports", "80,443")
        try:
            self.ports = self.helpers.parse_port_string(str(port_string))
        except ValueError as e:
            return False, f"Error parsing ports: {e}"

        if not self.portscanner_enabled:
            self.info(f"No portscanner enabled. Assuming open ports: {', '.join(str(x) for x in self.ports)}")

        target_len = len(self.scan.target.seeds)
        if target_len > self.config.get("max_hosts", 65536):
            if not self.portscanner_enabled:
                self.hugewarning(
                    f"Selected target ({target_len:,} hosts) is too large, skipping IP_RANGE --> IP_ADDRESS speculation"
                )
                self.hugewarning('Enabling the "portscan" module is highly recommended')
            self.range_to_ip = False

        return True

    async def handle_event(self, event):
        ### BEGIN ESSENTIAL SPECULATION ###
        # These features are required for smooth operation of bbot
        # I.e. they are not "osinty" or intended to discover anything, they only compliment other modules

        # we speculate on distance-1 stuff too, because distance-1 open ports are needed by certain modules like sslcert
        event_in_scope_distance = event.scope_distance <= (self.scan.scope_search_distance + 1)
        speculate_open_ports = self.emit_open_ports and event_in_scope_distance

        # generate individual IP addresses from IP range
        if event.type == "IP_RANGE" and self.range_to_ip:
            net = ipaddress.ip_network(event.data)
            ips = list(net)
            random.shuffle(ips)
            for ip in ips:
                await self.emit_event(
                    ip,
                    "IP_ADDRESS",
                    parent=event,
                    internal=True,
                    context=f"speculate converted range into individual IP_ADDRESS: {ip}",
                )

        # IP_ADDRESS / DNS_NAME --> OPEN_TCP_PORT
        if speculate_open_ports:
            # don't act on unresolved DNS_NAMEs
            usable_dns = False
            if event.type == "DNS_NAME":
                if self.dns_disable or event.resolved_hosts:
                    usable_dns = True

            if event.type == "IP_ADDRESS" or usable_dns:
                for port in self.ports:
                    await self.emit_event(
                        self.helpers.make_netloc(event.data, port),
                        "OPEN_TCP_PORT",
                        parent=event,
                        internal=True,
                        context="speculated {event.type}: {event.data}",
                    )

        ### END ESSENTIAL SPECULATION ###
        if self.essential_only:
            return

        # parent domains
        if event.type.startswith("DNS_NAME"):
            parent = self.helpers.parent_domain(event.host_original)
            if parent != event.data:
                await self.emit_event(
                    parent, "DNS_NAME", parent=event, context="speculated parent {event.type}: {event.data}"
                )

        # URL --> OPEN_TCP_PORT
        event_is_url = event.type == "URL"
        if event_is_url or (event.type == "URL_UNVERIFIED" and self.open_port_consumers):
            # only speculate port from a URL if it wouldn't be speculated naturally from the host
            if event.host and (event.port not in self.ports or not speculate_open_ports):
                await self.emit_event(
                    self.helpers.make_netloc(event.host, event.port),
                    "OPEN_TCP_PORT",
                    parent=event,
                    internal=not event_is_url,  # if the URL is verified, the port is definitely open
                    context=f"speculated {{event.type}} from {event.type}: {{event.data}}",
                )

        # speculate sub-directory URLS from URLS
        if event.type == "URL":
            url_parents = self.helpers.url_parents(event.data)
            for up in url_parents:
                url_event = self.make_event(f"{up}/", "URL_UNVERIFIED", parent=event)
                if url_event is not None:
                    # inherit web spider distance from parent (don't increment)
                    parent_web_spider_distance = getattr(event, "web_spider_distance", 0)
                    url_event.web_spider_distance = parent_web_spider_distance
                    await self.emit_event(url_event, context="speculated web sub-directory {event.type}: {event.data}")

        # speculate URL_UNVERIFIED from URL or any event with "url" attribute
        event_is_url = event.type == "URL"
        event_has_url = isinstance(event.data, dict) and "url" in event.data
        event_tags = ["httpx-safe"] if event.type in ("CODE_REPOSITORY", "SOCIAL") else []
        if event_is_url or event_has_url:
            if event_is_url:
                url = event.data
            else:
                url = event.data["url"]
            # only emit the url if it's not already in the event's history
            if not any(e.type == "URL_UNVERIFIED" and e.data == url for e in event.get_parents()):
                await self.emit_event(
                    url,
                    "URL_UNVERIFIED",
                    tags=event_tags,
                    parent=event,
                    context="speculated {event.type}: {event.data}",
                )

        # ORG_STUB from TLD, SOCIAL, AZURE_TENANT
        org_stubs = set()
        if event.type == "DNS_NAME" and event.scope_distance == 0:
            tldextracted = self.helpers.tldextract(event.data)
            registered_domain = getattr(tldextracted, "registered_domain", "")
            if registered_domain:
                tld_stub = getattr(tldextracted, "domain", "")
                if tld_stub:
                    decoded_tld_stub = self.helpers.smart_decode_punycode(tld_stub)
                    org_stubs.add(decoded_tld_stub)
                    org_stubs.add(self.helpers.unidecode(decoded_tld_stub))
        elif event.type == "SOCIAL":
            stub = event.data.get("stub", "")
            if stub:
                org_stubs.add(stub.lower())
        elif event.type == "AZURE_TENANT":
            tenant_names = event.data.get("tenant-names", [])
            org_stubs.update(set(tenant_names))
        for stub in org_stubs:
            stub_hash = hash(stub)
            if stub_hash not in self.org_stubs_seen:
                self.org_stubs_seen.add(stub_hash)
                stub_event = self.make_event(stub, "ORG_STUB", parent=event)
                if stub_event:
                    await self.emit_event(stub_event, context="speculated {event.type}: {event.data}")

        # USERNAME --> EMAIL
        if event.type == "USERNAME":
            email = event.data.split(":", 1)[-1]
            if validators.soft_validate(email, "email"):
                email_event = self.make_event(email, "EMAIL_ADDRESS", parent=event, tags=["affiliate"])
                if email_event:
                    await self.emit_event(email_event, context="detected {event.type}: {event.data}")
